# -*- coding: utf-8 -*-
"""
Created on Mon Nov 20 10:16:13 2017

@author: erwan

Examples
--------

Run all tests::

    pytest       (in command line, in project folder)

Run only fast tests (i.e: tests that a 'fast' label)::

    pytest -m fast


-------------------------------------------------------------------------------

"""

import os
from os.path import basename, exists, getmtime

import numpy as np
import pytest
from numpy import exp

from radis import SpectrumFactory
from radis.db.molecules import Molecules
from radis.levels.partfunc import PartFunc_Dunham, PartFuncKurucz, PartFuncTIPS
from radis.levels.partfunc_cdsd import PartFuncCO2_CDSDcalc, PartFuncCO2_CDSDtab
from radis.misc.printer import printm
from radis.misc.warning import DeprecatedFileWarning
from radis.phys.constants import hc_k
from radis.test.utils import getTestFile, setup_test_line_databases

fig_prefix = basename(__file__) + ": "


from radis.misc.utils import NotInstalled, not_installed_vaex_args

try:
    import vaex
except ImportError:
    vaex = NotInstalled(*not_installed_vaex_args)

# %% Test

# never add @pytest.mark.fast so we don't delete cached files for 'fast' tests
def test_delete_all_cached_energies(verbose=True, warnings=True, *args, **kwargs):
    """Doesnt really test anything, but cleans all cached energy levels"""

    for molecule, isotopes in Molecules.items():
        for isotope, states in isotopes.items():
            for state, ElecState in states.items():
                # Delete cache file for Dunham expansions
                energies = PartFunc_Dunham(ElecState, use_cached="return")
                cachefile = energies.cachefile
                if exists(cachefile):
                    os.remove(cachefile)
                    print(
                        (
                            "Cleaned cached energies for {0}".format(
                                ElecState.get_fullname()
                            )
                        )
                    )


@pytest.mark.fast
def test_cache_file_generation_and_update(verbose=True, *args, **kwargs):
    """Test that cache file process works correctly, using CO as an example.

    Expected behavior:

    - generated when 'regen'
    - updated when using 'True' but input parameters have changed
    - an error is raised when 'force' is used but input parameters have changed

    """
    # TODO: move it as a test of cache_files.py in RADIS

    S = Molecules["CO"][1]["X"]

    from os.path import exists, getmtime

    if verbose:
        printm("Testing behavior for Energy cache file with use_cached=")

    # Create a first cache file (or use existing one)
    db = PartFunc_Dunham(
        S, vmax=12, vmax_morse=48, Jmax=300, use_cached=True, verbose=verbose
    )

    # .. test file was created
    cachefile = db.cachefile
    assert exists(cachefile)
    last_modif = getmtime(cachefile)
    if verbose:
        printm("... True: energy cache file has been used/created (vmax=12)")

    # Force recalculating (using same inputs, but use_cached='regen')
    db = PartFunc_Dunham(
        S, vmax=12, vmax_morse=48, Jmax=300, use_cached="regen", verbose=verbose
    )

    # ... test that cache file was updated
    cachefile = db.cachefile
    assert getmtime(cachefile) > last_modif
    last_modif = getmtime(cachefile)
    if verbose:
        print("... 'regen': energy cache file has been created again (vmax=12)")

    # Recompute with different parameters

    # ... if using 'force', ensures that an error is raised
    with pytest.raises(DeprecatedFileWarning):  # DeprecatedFileError is expected
        db = PartFunc_Dunham(
            S, vmax=11, vmax_morse=48, Jmax=300, use_cached="force", verbose=verbose
        )
    if verbose:
        printm(
            "... 'force': an error was correctly triggered for new input parameters (vmax=11)"
        )

    # ... if using 'True', ensures that cache file is updated
    db = PartFunc_Dunham(
        S, vmax=11, vmax_morse=48, Jmax=300, use_cached=True, verbose=verbose
    )
    assert getmtime(cachefile) > last_modif
    if verbose:
        printm("... True: cache file was updated for new input parameters (vmax=11)")


@pytest.mark.needs_config_file
@pytest.mark.fast
@pytest.mark.needs_db_CDSD_HITEMP_PC
def test_CDSD_calc_vs_tab(verbose=True, warnings=True, *args, **kwargs):
    """Test 1: compare calculated PartFunc to the tabulated one"""

    from radis.misc.config import getDatabankEntries

    iso = 1
    database = "CDSD-HITEMP-PC"

    # Compare tab to hardcoded
    parfunc = getDatabankEntries(database)["parfunc"]
    Qf = PartFuncCO2_CDSDtab(iso, parfunc)
    assert np.isclose(Qf.at(300), 291.0447781984652, rtol=0.001)
    assert np.isclose(Qf.at(3000), 114689.88454184022, rtol=0.001)

    if verbose:
        printm("Tested CDSD tabulated is correct: OK")

    # Compare tab to calculated
    energies = getDatabankEntries(database)["levels"]
    levelsfmt = getDatabankEntries(database)["levelsfmt"]
    Qfc = PartFuncCO2_CDSDcalc(
        energies[iso], levelsfmt=levelsfmt, isotope=iso, use_cached=True
    )
    assert np.isclose(Qf.at(300), Qfc.at(300), rtol=0.001)
    assert np.isclose(Qf.at(3000), Qfc.at(3000), rtol=0.001)

    if verbose:
        printm("Tested CDSD Q_calc vs Q_tab give same output: OK")

    return True


@pytest.mark.needs_config_file
@pytest.mark.fast
def test_reduced_CDSD_calc_vs_tab(verbose=True, warnings=True, *args, **kwargs):
    """Test 1: compare calculated PartFunc to the tabulated one

    Version where we use the reduced set of CO2 levels (< 3000 cm-1)"""
    from radis import setup_test_line_databases
    from radis.misc.config import getDatabankEntries

    setup_test_line_databases()  # needed for "HITEMP-CO2-HAMIL-TEST"

    iso = 1
    database = "HITEMP-CO2-HAMIL-TEST"

    # Compare tab to hardcoded
    parfunc = getDatabankEntries(database)["parfunc"]
    Qf = PartFuncCO2_CDSDtab(iso, parfunc)
    assert np.isclose(Qf.at(300), 291.0447781984652, rtol=0.001)
    assert np.isclose(Qf.at(3000), 114689.88454184022, rtol=0.001)

    if verbose:
        printm("Tested CDSD tabulated is correct: OK")

    # Compare tab to calculated
    energies = getDatabankEntries(database)["levels"]
    levelsfmt = getDatabankEntries(database)["levelsfmt"]
    Qfc = PartFuncCO2_CDSDcalc(
        energies[iso],
        levelsfmt=levelsfmt,
        isotope=iso,
        use_cached=True,
        verbose=verbose,
    )
    assert np.isclose(
        Qf.at(300), Qfc.at(300), rtol=0.01
    )  # reduced rtol to accommodate for the reduced set of levels in the test set
    # assert np.isclose(Qf.at(3000), Qfc.at(3000), rtol=0.001)  # of course doesnt work with the reduced set of levels in the test set

    if verbose:
        printm("Tested CDSD Q_calc vs Q_tab give same output: OK")


# @pytest.mark.fast #not fast, Nicolas Minesi 08/04/2024
def test_calculatedQ_match_HAPI_CO(
    vmax=11, jmax=300, plot=False, verbose=True, *args, **kwargs
):
    """Tested that Q ab_initio (Dunham) match HAPI for CO at different temperatures"""

    vmax = 11
    vmax_morse = 48
    jmax = 300
    iso = 1

    S = Molecules["CO"][iso]["X"]

    # Dont use cached: force recalculating
    db = PartFunc_Dunham(
        S, vmax=vmax, vmax_morse=vmax_morse, Jmax=jmax, use_cached=False
    )  # , ZPE=1081.584383)
    assert not db.use_cached

    #    if plot: db.plot_states()

    hapi = PartFuncTIPS(M=5, I=1)  # CO  # isotope

    us = []
    hap = []
    T = np.linspace(300, 3000)
    for Ti in T:
        us.append(db.at(Ti))
        hap.append(hapi.at(Ti))

    if plot:  # Make sure matplotlib is interactive so that test are not stuck in pytest
        import matplotlib.pyplot as plt

        plt.ion()

        plt.figure(fig_prefix + "Partition function Dunham vs Precomputed")
        plt.plot(T, us, "ok", label="NeQ")
        plt.plot(T, hap, "or", label="HAPI")
        plt.legend()
        plt.xlabel("Temperature (K)")
        plt.ylabel("Partition function")
        plt.title(
            "Ab-initio partition function calculations\n"
            + "(vmax:{0},jmax:{1})".format(vmax, jmax)
        )
        plt.tight_layout()

    # Compare Calculated  vs HAPI
    assert np.allclose(us, hap, rtol=0.02)

    if verbose:
        printm("Tested Q_CO ab_initio (Dunham) matches HAPI for 300 - 3000 K: OK")

    return True


# @pytest.mark.fast #not fast, Nicolas Minesi 08/04/2024
def test_calculatedQ_match_HAPI(plot=False, verbose=True, *args, **kwargs):
    """Tested that Q ab_initio (Dunham) match HAPI for different molecules
    and isotopes"""

    # molecule, isotope, temperature, absolute tolerance
    for molecule, iso, T, atol, rtol in [
        ("CO2", 1, 300, 0.9, 0.02),
        ("CO2", 1, 1000, 5.0, 0.02),
        ("CO2", 1, 3000, 2150, 0.02),
        ("CO2", 2, 300, 1.8, 0.02),
        ("CO2", 2, 1000, 25, 0.02),
        ("CO2", 2, 3000, 4962, 0.02),
        ("CO", 1, 300, 0.16, 0.02),
        ("CO", 1, 1000, 1.9, 0.02),
        ("CO", 1, 3000, 27, 0.02),
        ("CO", 2, 300, 0.31, 0.02),
        ("CO", 2, 1000, 4.1, 0.02),
        ("CO", 2, 3000, 56.9, 0.02),
        ("CO", 3, 300, 0.16, 0.02),
        ("CO", 3, 1000, 2.1, 0.02),
        ("CO", 3, 3000, 28.6, 0.02),
    ]:

        S = Molecules[molecule][iso]["X"]

        # Dont use cached: force recalculating
        db = PartFunc_Dunham(S)

        from radis.db.classes import get_molecule_identifier

        hapi = PartFuncTIPS(M=get_molecule_identifier(molecule), I=iso)

        Q_radis = db.at(T)
        Q_hapi = hapi.at(T)

        if verbose:
            print(
                "Q({0},iso={1},{2}K)\tRADIS: {3:.2f}\tHAPI={4:.2f}".format(
                    molecule, iso, T, Q_radis, Q_hapi
                )
            )

        try:
            assert np.isclose(Q_radis, Q_hapi, atol=atol)
            assert np.isclose(Q_radis, Q_hapi, atol=atol)
        except AssertionError:
            raise AssertionError(
                "Partition function for {0}, iso={1} ".format(molecule, iso)
                + "at {0}K doesnt match in RADIS ".format(T)
                + "({0:.4f}) and HAPI ({1:.4f})".format(Q_radis, Q_hapi)
            )

    if verbose:
        printm("Tested Q ab_initio (Dunham) matches HAPI: OK")

    return True


@pytest.mark.needs_config_file
@pytest.mark.fast
@pytest.mark.needs_db_CDSD_HITEMP_PC
def test_CDSD_calc_vs_ref(warnings=True, verbose=True, *args, **kwargs):
    """Test partition functions calculated with CDSD energy levels against
    hardcoded values"""

    from radis.misc.config import getDatabankEntries

    iso = 1

    energies = getDatabankEntries("CDSD-HITEMP-PC")["levels"]
    levelsfmt = getDatabankEntries("CDSD-HITEMP-PC")["levelsfmt"]
    Qf = PartFuncCO2_CDSDcalc(
        energy_levels=energies[iso],
        isotope=iso,
        use_cached=True,
        levelsfmt=levelsfmt,
    )
    assert np.isclose(Qf.at(300), 291.0447781984652, rtol=0.001)
    assert np.isclose(Qf.at(3000), 114689.88454184022, rtol=0.001)
    assert np.isclose(Qf.at_noneq(300, 300), 291.0447781984652, rtol=0.001)
    assert np.isclose(Qf.at_noneq(3000, 3000), 114689.88454184022, rtol=0.001)
    assert np.isclose(
        Qf.at_noneq(3000, 3000, overpopulation={"(0,1)": 3}, returnQvibQrot=True)[0],
        120053.34252537244,
        rtol=0.001,
    )

    if verbose:
        printm("Tested Q_CDSD values are correct : OK")


def test_partfunckurucz():
    species = "Ca_I"
    kurucz_data = PartFuncKurucz(species)

    temperatures = np.array(
        [
            1e-05,
            1e-04,
            1e-03,
            1e-02,
            1e-01,
            1.5e-01,
            2e-01,
            3e-01,
            5e-01,
            7e-01,
            1.0,
            1.3,
            1.7,
            2.0,
            3.0,
            5.0,
            7.0,
            10.0,
            15.0,
            20.0,
            30.0,
            50.0,
            70.0,
            100.0,
            130.0,
            170.0,
            200.0,
            250.0,
            300.0,
            500.0,
            700.0,
            1e03,
            1.5e03,
            2e03,
            3e03,
            4e03,
            5e03,
            6e03,
            7e03,
            8e03,
            9e03,
            1e04,
        ]
    )
    pf_values = np.array(
        [1.0] * 33
        + [
            1.00016,
            1.00701,
            1.04991,
            1.17173,
            1.41809,
            1.85725,
            2.60365,
            3.81954,
            5.69578,
        ]
    )

    for T, expected_pf in zip(temperatures, pf_values):
        assert np.isclose(
            kurucz_data._at(T), expected_pf, atol=1e-4
        ), f"Failed at T={T}"

    T_test = 3000  # Temperature at which to test interpolation

    # Compute expected partition function value using np.interp directly on the original data
    expected_pf = np.interp(T_test, temperatures, pf_values)

    assert np.isclose(
        kurucz_data._at(T_test), expected_pf, atol=1e-4
    ), f"Interpolation failed at T={T_test}"


def test_reduced_CDSD_calc_noneq(verbose=True, warnings=True, *args, **kwargs):
    """Compare calculated partition function at equilibrium and nonequilibrium
    using the CDSD-format

    Examples
    --------

        assert Qfc.at(300) != Qfc.at_noneq_3Tvib((300, 300, 300), 300)

    After redefining Evib so that Evib + Erot = E:

        assert np.isclose(Qfc.at(300), Qfc.at_noneq(300, 300), rtol=0.001)
        assert np.isclose(Qfc.at(300), Qfc.at_noneq_3Tvib((300, 300, 300), 300), rtol=0.001)
    """

    from radis.misc.config import getDatabankEntries
    from radis.test.utils import define_Evib_as_sum_of_Evibi

    iso = 1
    database = "HITEMP-CO2-HAMIL-TEST"

    # Compare tab to calculated
    energies = getDatabankEntries(database)["levels"]
    levelsfmt = getDatabankEntries(database)["levelsfmt"]
    Qfc = PartFuncCO2_CDSDcalc(
        energies[iso],
        levelsfmt=levelsfmt,
        isotope=iso,
        use_cached=True,
        verbose=verbose,
    )

    # Note that partition functions are not equal because some of the energy
    # goes in coupling terms in the Hamiltonian formulation
    assert Qfc.at(300) != Qfc.at_noneq_3Tvib((300, 300, 300), 300)

    # Below we redefine Evib so that Evib + Erot = E
    define_Evib_as_sum_of_Evibi(Qfc.df)

    # New Test:
    assert np.isclose(Qfc.at(300), Qfc.at_noneq(300, 300), rtol=0.001)
    assert np.isclose(Qfc.at(300), Qfc.at_noneq_3Tvib((300, 300, 300), 300), rtol=0.001)


# @pytest.mark.fast   # (very fast only once the cached database has been generated, else decently fast)
def test_recompute_Q_from_QvibQrot_Dunham_Evib3_Evib12Erot(
    verbose=True, warnings=True, *args, **kwargs
):
    """Calculate vibrational and rotational partition functions:

    - with Dunham expansions. Evib, Erot = (Evib3, Evib1+Evib2+Erot)
    - under nonequilibrium

    Calculate total rovibrational partition function, and compare

    Test if partition function can be recomputed correctly from vibrational
    populations and rotational partition function (note that we are in a coupled
    case so partition function is not simply the product of Qvib, Qrot)
    """

    iso = 1
    Tvib = 1500
    Trot = 300

    S = Molecules["CO2"][iso]["X"]

    Qf = PartFunc_Dunham(
        S,
        use_cached=True,
        group_energy_modes_in_2T_model={"CO2": (["Evib3"], ["Evib1", "Evib2", "Erot"])},
    )

    Q = Qf.at_noneq(Tvib, Trot)
    _, Qvib, dfQrot = Qf.at_noneq(Tvib, Trot, returnQvibQrot=True)
    if verbose:
        printm("Q", Q)
    if verbose:
        printm("Qvib", Qvib)

    # 1) Test Q vs Q recomputed from Qrot, Qvib

    # Recompute Qtot
    df = dfQrot
    Q2 = ((df.gvib * exp(-df.Evib * hc_k / Tvib)) * df.Qrot).sum()
    # Todo: non Boltzmann case

    assert np.isclose(Q, Q2)

    if verbose:
        printm("Tested Q vs recomputed from (Qvib, Qrot) are the same: OK")

    return True


# @pytest.mark.fast   # (very fast only once the cached database has been generated, else decently fast)
def test_recompute_Q_from_QvibQrot_Dunham_Evib123_Erot(
    verbose=True, warnings=True, *args, **kwargs
):
    """Calculate vibrational and rotational partition functions:

    - with Dunham expansions. Evib, Erot = (Evib1+Evib2+Evib3, Erot)
    - under nonequilibrium

    Calculate total rovibrational partition function, and compare

    Test if partition function can be recomputed correctly from vibrational
    populations and rotational partition function (note that we are in a coupled
    case so partition function is not simply the product of Qvib, Qrot)
    """

    iso = 1
    Tvib = 1500
    Trot = 300

    S = Molecules["CO2"][iso]["X"]

    Qf = PartFunc_Dunham(
        S,
        use_cached=True,
        group_energy_modes_in_2T_model={"CO2": (["Evib1", "Evib2", "Evib3"], ["Erot"])},
    )

    Q = Qf.at_noneq(Tvib, Trot)
    _, Qvib, dfQrot = Qf.at_noneq(Tvib, Trot, returnQvibQrot=True)
    if verbose:
        printm("Q", Q)
    if verbose:
        printm("Qvib", Qvib)

    # 1) Test Q vs Q recomputed from Qrot, Qvib

    # Recompute Qtot
    df = dfQrot
    Q2 = ((df.gvib * exp(-df.Evib * hc_k / Tvib)) * df.Qrot).sum()
    # Todo: non Boltzmann case

    assert np.isclose(Q, Q2)

    if verbose:
        printm("Tested Q vs recomputed from (Qvib, Qrot) are the same: OK")

    return True


# @pytest.mark.needs_config_file
# @pytest.mark.fast
# @pytest.mark.needs_db_CDSD_HITEMP
# def test_recompute_Q_from_QvibQrot_CDSD_PCN(verbose=True, warnings=True, *args, **kwargs):
#    '''
#    Calculate vibrational and rotational partition functions:
#
#    - in CDSD with (p,c,N) convention for vibrational levels
#    - under nonequilibrium
#
#    Calculate total rovibrational partition function, and compare
#
#    Test if partition function can be recomputed correctly from vibrational
#    populations and rotational partition function (note that we are in a coupled
#    case so partition function is not simply the product of Qvib, Qrot)
#    '''
#
#    from radis.misc.config import getDatabankEntries
#
#    iso = 1
#
#    try:
#        energies = getDatabankEntries('CDSD-HITEMP-PCN')['levels']
#        levelsfmt = getDatabankEntries('CDSD-HITEMP-PCN')['levelsfmt']
#
#        Tvib = 1500
#        Trot = 300
#
#        Qf = PartFuncCO2_CDSDcalc(energies[iso],
#                                  isotope=iso,
#                                  use_cached=True,
#                                  levelsfmt=levelsfmt)
#        Q = Qf.at_noneq(Tvib, Trot)
#        _, Qvib, dfQrot = Qf.at_noneq(Tvib, Trot, returnQvibQrot=True)
#        if verbose:
#            printm('Q', Q)
#        if verbose:
#            printm('Qvib', Qvib)
#
#        # 1) Test Q vs Q recomputed from Qrot, Qvib
#
#        # Recompute Qtot
#        df = dfQrot
#        Q2 = ((df.gvib*exp(-df.Evib*hc_k/Tvib))*df.Qrot).sum()
#        # Todo: non Boltzmann case
#
#        assert np.isclose(Q, Q2)
#
#        if verbose:
#            printm('Tested Q vs recomputed from (Qvib, Qrot) are the same: OK')
#
#        return True
#
#    except DatabankNotFound as err:
#        assert IgnoreMissingDatabase(err, __file__, warnings)


@pytest.mark.needs_config_file
@pytest.mark.fast
@pytest.mark.needs_db_CDSD_HITEMP
def test_recompute_Q_from_QvibQrot_CDSD_PC(
    verbose=True, warnings=True, *args, **kwargs
):
    """
    Calculate vibrational and rotational partition functions:

    - in CDSD with (p,c) convention for vibrational levels
    - under nonequilibrium

    Recompute total partition function, and compare

    Test if partition function can be recomputed correctly from vibrational
    populations and rotational partition function (note that we are in a coupled
    case so partition function is not simply the product of Qvib, Qrot)
    """

    from radis.misc.config import getDatabankEntries

    iso = 1

    energies = getDatabankEntries("CDSD-HITEMP-PC")["levels"]
    levelsfmt = getDatabankEntries("CDSD-HITEMP-PC")["levelsfmt"]

    Tvib = 1500
    Trot = 300

    Qf = PartFuncCO2_CDSDcalc(
        energies[iso], isotope=iso, use_cached=True, levelsfmt=levelsfmt
    )
    Q = Qf.at_noneq(Tvib, Trot)
    _, Qvib, dfQrot = Qf.at_noneq(Tvib, Trot, returnQvibQrot=True)
    if verbose:
        printm("Q", Q)
    if verbose:
        printm("Qvib", Qvib)

    # 1) Test Q vs Q recomputed from Qrot, Qvib

    # Recompute Qtot
    df = dfQrot
    Q2 = ((df.gvib * exp(-df.Evib * hc_k / Tvib)) * df.Qrot).sum()
    # Todo: non Boltzmann case

    assert np.isclose(Q, Q2)

    if verbose:
        printm("Tested Q vs recomputed from (Qvib, Qrot) are the same: OK")

    return True


# @pytest.mark.fast            # (fast only once the cached database has been generated)


def test_Q_1Tvib_vs_Q_3Tvib(T=1500, verbose=True, warnings=True, *args, **kwargs):
    """Test if partition function calculated in 1-Tvib mode returns the same
    result as partition function calculated in 3-Tvib mode
    """

    b = True

    # input
    M = "CO2"
    I = 1  # isotope
    S = Molecules[M][I]["X"]

    Qf = PartFunc_Dunham(S, use_cached=True)
    df = Qf.df

    # First make sure energies match
    if not (df.Evib == df.Evib1 + df.Evib2 + df.Evib3).all():
        b *= False
        if warnings:
            printm("WARNING in test_Q_1Tvib_vs_Q_3Tvib: Evib != Evib1 + Evib2 + Evib3")

    # Then calculate Q vs Q3T
    Q = Qf.at_noneq(T, T)
    if verbose:
        printm("Q", Q)

    Q3T = Qf.at_noneq_3Tvib((T, T, T), T)
    if verbose:
        printm("Q3T", Q3T)

    assert np.isclose(Q, Q3T)
    if verbose:
        printm("Tested Q in 1-Tvib vs Q in 3-Tvib modes (T={0:.0f}K): OK".format(T))

    return True


@pytest.mark.fast
def test_Morse_Potential_effect_CO(
    T=3000, rtol=1e-4, verbose=True, warnings=True, *args, **kwargs
):
    """Quantify effect of calculating upper levels near dissociation limit
    with Morse Potential

    Returns True if difference is less than rtol
    """

    vmax = 11
    vmax_morse = 48
    jmax = 300
    iso = 1
    S = Molecules["CO"][iso]["X"]

    # TODO: remove vmax, vmax_morse from Dunham method. Only use the one in ElecState
    db = PartFunc_Dunham(S, vmax=vmax, vmax_morse=0, Jmax=jmax, use_cached=False)
    Q_nomorse = db.at(T)

    db = PartFunc_Dunham(
        S, vmax=vmax, vmax_morse=vmax_morse, Jmax=jmax, use_cached=False
    )
    Q_morse = db.at(T)

    if verbose:
        printm("Morse vs no Morse potential (T={0}K)".format(T))
        printm("Q_morse: {0:.3f}".format(Q_morse))
        printm("Q_nomorse: {0:.3f}".format(Q_nomorse))
        printm("Difference: {0:.4f}%".format(abs(Q_nomorse - Q_morse) / Q_morse * 100))

    assert abs(Q_nomorse - Q_morse) / Q_morse < rtol


def test_levels_regeneration(verbose=True, warnings=True, *args, **kwargs):
    """Test that we regenerate levels file automatically if manually changed

    see https://github.com/radis/radis/issues/90

    """

    # from warnings import catch_warnings, filterwarnings
    def run_example():

        from radis.test.utils import (
            define_Evib_as_sum_of_Evibi,
            discard_lines_with_na_levels,
        )

        setup_test_line_databases(
            verbose=True
        )  # add HITEMP-CO2-HAMIL-TEST in ~/radis.json if not there

        sf = SpectrumFactory(
            wavenum_min=2283.7,
            wavenum_max=2285.1,
            wstep=0.001,
            cutoff=1e-30,
            path_length=0.1,
            mole_fraction=400e-6,
            isotope=[1],
            verbose=2,
        )
        sf.warnings["MissingSelfBroadeningWarning"] = "ignore"
        sf.load_databank(
            "HITEMP-CO2-HAMIL-TEST",
            load_energies=True,  # Loading Energy Database
            db_use_cached=True,  # important to test Cache file here
        )

        # Now generate vibrational energies for a 2-T model
        levels = sf.parsum_calc["CO2"][1]["X"].df
        define_Evib_as_sum_of_Evibi(levels)
        discard_lines_with_na_levels(sf)

        # Calculate populations using the non-equilibrium module:
        # This will crash the first time because the Levels Database is just a fragment and does not include all levels.
        try:
            sf.non_eq_spectrum(300, 300)
        except AssertionError:  # expected
            sf.df0.dropna(inplace=True)

        s = sf.non_eq_spectrum(300, 300)
        s.plot()

    # run calculation (SpectrumFactory)
    run_example()
    # get the time when .levels file was last modified
    levels_last_modification = getmtime(
        getTestFile(r"co2_cdsd_hamiltonian_fragment.levels")
    )
    # get the time when .levels.h5 file was last modified
    cache_last_modification = getmtime(
        getTestFile(r"co2_cdsd_hamiltonian_fragment.levels.h5")
    )

    # change the time when .levels file was last modified
    stinfo = os.stat(getTestFile(r"co2_cdsd_hamiltonian_fragment.levels"))
    access_time = stinfo.st_atime
    os.utime(
        getTestFile(r"co2_cdsd_hamiltonian_fragment.levels"),
        (levels_last_modification + 1, access_time + 1),
    )

    # check if the change was successful
    levels_last_modification_again = getmtime(
        getTestFile(r"co2_cdsd_hamiltonian_fragment.levels")
    )

    assert levels_last_modification_again > levels_last_modification

    # run calculations once again to see if the .levels.h5 (cache file) is regenerated
    run_example()

    cache_last_modification_again = getmtime(
        getTestFile(r"co2_cdsd_hamiltonian_fragment.levels.h5")
    )
    assert cache_last_modification_again > cache_last_modification


@pytest.mark.skipif(
    isinstance(vaex, NotInstalled),
    reason="Vaex not available, just-in-time partition functions are only implemented in Vaex as of Radis 0.15",
)
def test_tabulated_partition_functions(
    verbose=True, plot=True, rtol=1e-2, *args, **kwargs
):
    """Test just-in-time tabulated partition return the same results as
    full summation within 0.5%  (adjust value with ``rtol``)

    We compute up to 10,000 K (beyond the validity range of most spectroscopic
    parameters)

    Run with verbose=True to check the accuracy and calibrate the
    :py:attr:`radis.levels.partfunc.RovibParFuncCalculator.N_bins_scaling` function"""

    from radis.db.molecules import CO2_X_626
    from radis.levels.partfunc import PartFunc_Dunham, PartFuncHAPI

    Z_sum = PartFunc_Dunham(CO2_X_626, mode="full summation")
    Z_tab = PartFunc_Dunham(
        CO2_X_626, mode="tabulation", verbose=3 if verbose else False
    )

    # For reference, start by comparing full-summation partition functions with TIPS tabulated partition functions
    Z_tips = PartFuncHAPI(M=2, I=1)  # CO  # isotope
    for T in [296, 3000, 5000]:
        z1 = Z_sum.at(T)
        z2 = Z_tips.at(T)
        accuracy = max(z1, z2) / min(z1, z2) - 1
        # accuracy_dict[
        if verbose:
            print(
                "full-sum & TIPS tabulated partition function of CO2 at {0}K close within {1:.2%}".format(
                    T, accuracy
                )
            )
        assert (
            accuracy < 4 * rtol
        )  # doesnt make sense to be extremely accurate in our tabulation if the full-summation doesnt match TIPS anyway

    # accuracy_dict = {1:[], 2:[], 3:[]}  # accuracy as a function of number of temperatures (helps how to scale N_bins)

    def compare_partition_functions(Z1, Z2, *T):
        z1 = Z1(*T)
        z2 = Z2(*T)
        # accuracy_dict[
        if verbose:
            accuracy = max(z1, z2) / min(z1, z2) - 1
            print(
                "full-sum & jit-tabulated partition function of CO2 at {0}K close within {1:.2%}".format(
                    T, accuracy
                )
            )
        assert np.isclose(z1, z2, rtol=rtol)

    # Equilibrium (same < 0.2%)
    compare_partition_functions(Z_sum.at, Z_tab.at, 296)
    compare_partition_functions(Z_sum.at, Z_tab.at, 3000)
    compare_partition_functions(Z_sum.at, Z_tab.at, 5000)
    compare_partition_functions(Z_sum.at, Z_tab.at, 10000)

    # Nonequilibrium (same << 0.1%)

    #  ... Compare with Partition function computed from PartFunc_Dunham
    compare_partition_functions(Z_sum.at_noneq, Z_tab.at_noneq, 296, 296)
    compare_partition_functions(Z_sum.at_noneq, Z_tab.at_noneq, 1000, 300)
    compare_partition_functions(Z_sum.at_noneq, Z_tab.at_noneq, 1000, 3000)
    compare_partition_functions(Z_sum.at_noneq, Z_tab.at_noneq, 3000, 3000)
    compare_partition_functions(Z_sum.at_noneq, Z_tab.at_noneq, 5000, 5000)
    compare_partition_functions(Z_sum.at_noneq, Z_tab.at_noneq, 10000, 10000)

    # Nonequilibrium 3 Tvib (same << 0.1%)

    #  ... Compare with Partition function computed from PartFunc_Dunham
    compare_partition_functions(
        Z_sum.at_noneq_3Tvib, Z_tab.at_noneq_3Tvib, (296, 296, 296), 296
    )
    compare_partition_functions(
        Z_sum.at_noneq_3Tvib, Z_tab.at_noneq_3Tvib, (1000, 1000, 2000), 300
    )
    compare_partition_functions(
        Z_sum.at_noneq_3Tvib, Z_tab.at_noneq_3Tvib, (1000, 1000, 3500), 3000
    )
    compare_partition_functions(
        Z_sum.at_noneq_3Tvib, Z_tab.at_noneq_3Tvib, (3000, 3000, 3000), 3000
    )
    compare_partition_functions(
        Z_sum.at_noneq_3Tvib, Z_tab.at_noneq_3Tvib, (5000, 5000, 5000), 5000
    )
    compare_partition_functions(
        Z_sum.at_noneq_3Tvib, Z_tab.at_noneq_3Tvib, (10000, 10000, 10000), 10000
    )

    # ... change Grid :
    if plot:  # Make sure matplotlib is interactive so that test are not stuck in pytest
        import matplotlib.pyplot as plt

        plt.ion()

        Tvib = 3000
        Trot_arr = np.linspace(300, 3000, 10)
        plt.figure()
        plt.plot(Trot_arr, [Z_sum.at_noneq(Tvib, T) for T in Trot_arr])
        for N_bins in [10, 100, 1000]:
            Z_tab.N_bins == N_bins
            plt.plot(
                Trot_arr,
                [Z_tab.at_noneq(Tvib, T) for T in Trot_arr],
                "--",
                label=N_bins,
            )
        plt.legend()


@pytest.mark.skipif(
    isinstance(vaex, NotInstalled),
    reason="Vaex not available, just-in-time partition functions are only implemented in Vaex as of Radis 0.15",
)
def test_parsum_mode_in_factory(verbose=True, plot=True, *args, **kwargs):
    """Test Partition function modes in SpectrumFactory

    using :py:meth:`~radis.spectrum.spectrum.Spectrum.print_perf_profile`

    ::
        # sf.params.parsum_mode = 'full summation'   # default

        full summation profiler :
            spectrum_calculation      0.197s ████████████████
                check_line_databank             0.002s
                check_non_eq_param              0.000s
                reinitialize                    0.006s
                    copy_database                   0.003s
                    memory_usage_warning            0.003s
                    reset_population                0.000s
                calc_noneq_population           0.119s █████████
                    part_function                   0.020s █
                    population                      0.099s ████████
                scaled_non_eq_linestrength      0.002s
                    map_part_func                   0.000s
                    corrected_population_se         0.002s
                calc_emission_integral          0.012s
                calc_lineshift                  0.000s
                calc_hwhm                       0.013s █
                generate_wavenumber_arrays      0.000s
                calc_line_broadening            0.040s ███
                    precompute_LDM_lineshapes       0.002s
                    LDM_Initialized_vectors         0.000s
                    LDM_closest_matching_line       0.000s
                    LDM_Distribute_lines            0.001s
                    LDM_convolve                    0.037s ██
                calc_other_spectral_quan        0.004s
                generate_spectrum_obj           0.000s

    ::
        # sf.params.parsum_mode = 'tabulation'

        tabulation profiler :
            spectrum_calculation      0.104s ████████████████
                check_line_databank             0.002s
                check_non_eq_param              0.001s
                reinitialize                    0.004s
                    copy_database                   0.001s
                    memory_usage_warning            0.003s
                    reset_population                0.000s
                calc_noneq_population           0.026s ███
                    part_function                   0.019s ██
                    population                      0.007s █
                scaled_non_eq_linestrength      0.003s
                    map_part_func                   0.000s
                    corrected_population_se         0.003s
                calc_emission_integral          0.010s █
                calc_lineshift                  0.001s
                calc_hwhm                       0.017s ██
                generate_wavenumber_arrays      0.001s
                calc_line_broadening            0.035s █████
                    precompute_LDM_lineshapes       0.001s
                    LDM_Initialized_vectors         0.000s
                    LDM_closest_matching_line       0.001s
                    LDM_Distribute_lines            0.000s
                    LDM_convolve                    0.032s ████
                    others                          0.001s
                calc_other_spectral_quan        0.003s
                generate_spectrum_obj           0.000s
                others                          0.002s

    So calculation of populations is ~5x faster and the spectrum calculation itself
    is 2x faster.

    """
    from radis import SpectrumFactory

    wmin, wmax = 2284, 2285
    sf = SpectrumFactory(
        wavenum_min=wmin,
        wavenum_max=wmax,
        molecule="CO2",
        isotope="1",
        truncation=2.5,  # cm-1
        medium="air",
        path_length=10.32,  # cm
        wstep=0.001,
        verbose=0,
    )
    conditions = {
        "mole_fraction": 0.2,
        "pressure": 1,
        "Ttrans": 2000,
        "Trot": 2100,
        "Tvib": 1500,
    }

    # ... initialize tests :
    sf.params.parsum_mode = "full summation"  # default
    sf.load_databank("HITEMP-CO2-TEST")
    sf._add_bands()
    _ = sf.non_eq_spectrum(**conditions)  # initialize energies, etc.
    # ... compare:
    s_HITEMP = sf.non_eq_spectrum(**conditions, name=sf.params.parsum_mode)
    s_HITEMP.print_perf_profile()

    # Now, again with tabulation:
    # ... init :
    sf.params.parsum_mode = "tabulation"
    sf.load_databank("HITEMP-CO2-TEST")
    sf._add_bands()
    _ = sf.non_eq_spectrum(**conditions)  # initialize energies, first tabulation
    # ... compare:
    s_HITEMP2 = sf.non_eq_spectrum(**conditions, name=sf.params.parsum_mode)
    s_HITEMP2.print_perf_profile()

    if plot:
        import matplotlib.pyplot as plt

        plt.ion()
        from radis import plot_diff

        plot_diff(s_HITEMP2, s_HITEMP, method="ratio")

    from radis import get_residual

    assert get_residual(s_HITEMP, s_HITEMP2, "abscoeff") < 6e-5


def _run_testcases(verbose=True, warnings=True, *args, **kwargs):

    # Test 0: delete all cached energies
    test_delete_all_cached_energies(verbose=verbose, warnings=warnings)

    # Test 1: test cache mechanism
    test_cache_file_generation_and_update(verbose=verbose, warnings=warnings)

    # Test 2: compare calculated PartFunc for CO to HAPI
    test_calculatedQ_match_HAPI_CO()
    # Test 2b: compare for many molecules and isotopes
    test_calculatedQ_match_HAPI()

    # Test 3: compare calculated PartFunc to the tabulated one with CDSD
    # Test 4: compare calculated PartFunc to hardcoded references
    test_CDSD_calc_vs_tab(verbose=verbose, warnings=warnings)
    test_CDSD_calc_vs_ref(warnings=warnings)
    test_reduced_CDSD_calc_vs_tab(verbose=verbose, warnings=warnings)
    test_reduced_CDSD_calc_noneq(verbose=verbose, warnings=warnings)

    # Test 5a, 5b: recompute Q from QvibQrot
    test_recompute_Q_from_QvibQrot_Dunham_Evib123_Erot(
        verbose=verbose, warnings=warnings
    )
    test_recompute_Q_from_QvibQrot_Dunham_Evib3_Evib12Erot(
        verbose=verbose, warnings=warnings
    )
    test_recompute_Q_from_QvibQrot_CDSD_PC(verbose=verbose, warnings=warnings)
    # test_recompute_Q_from_QvibQrot_CDSD_PCN(verbose=verbose, warnings=warnings)  # ignore in released version

    # Test 6:
    test_Q_1Tvib_vs_Q_3Tvib(verbose=verbose, warnings=warnings)

    # Test 7:
    test_Morse_Potential_effect_CO(verbose=verbose, warnings=warnings)

    # Test 8: Regenerates levels file if it's manually changed
    test_levels_regeneration(verbose=verbose, warnings=True, *args, **kwargs)

    # Test 9 : tabulation
    test_tabulated_partition_functions(verbose=verbose, *args, **kwargs)
    test_parsum_mode_in_factory(verbose=verbose, *args, **kwargs)

    # Test 10:
    test_partfunckurucz()

    return True


if __name__ == "__main__":
    # printm("Testing parfunc: {0}".format(_run_testcases()))
    printm("Testing partfunc.py:", pytest.main(["test_partfunc.py", "--pdb"]))



import os.path
from warnings import warn

import numpy as np
from scipy.constants import N_A, c, k
from scipy.fft import next_fast_len

from radis.gpu.params import (
    init_G_params,
    init_L_params,
    init_Q,
    set_G_params,
    set_L_params,
    set_pTQ,
)
from radis.gpu.structs import initData_t, iterData_t
from radis.misc.utils import getProjectRoot
from radis.misc.warning import NoGPUWarning

gpu_mod = None
init_h = initData_t()
iter_h = iterData_t()


def gpu_init(
    vmin,
    Nv,
    dv,
    dxG,
    dxL,
    v0,
    da,
    na,
    S0,
    El,
    gamma_arr,
    iso,
    Mm_arr,
    Q_intp_list,
    verbose=0,
    backend="gpu-cuda",
):
    """
    Initialize GPU-based calculation for emission and absorption spectra in spectroscopy.

    Parameters
    ----------
    vmin : float
        Minimum value frequency/wavenumber axis
    Nv : int
        Total number of frequency/wavenumber points.
    dv : float
        Stepsize of frequency/wavenumber axis (called wstep elsewhere in RADIS).
    dxG : float
        Relative grid spacing for Gaussian lineshapes.
    dxL : float
        Relative grid spacing for Lorentzian lineshapes.
    v0 : numpy.ndarray[np.float32]
        Array of line center frequencies (in cm-1).
    da : numpy.ndarray[np.float32]
        Pressure shift  (in cm-1.atm-1).
    na : numpy.ndarray[np.float32]
        Temperature dependency of Lorentzian widths
    S0 : numpy.ndarray[np.float32]
        Line intensity scaling factors.
    El : numpy.ndarray[np.float32]
        Lower level energy levels.
    gamma_arr : numpy.ndarray[np.float32]
        (m,n) shaped array with Lorentzian width parameters, with n the number of lines in the database
        and m the number of collision partners included. This is usually at least two,
        with the first (m=0) always self broadening and the last (m=-1) always air broadening.
    iso : numpy.ndarray[np.uint8]
        Index of isotopologue.
    Mm_arr : numpy.ndarray
        Molecular masses for all isotopologues in the database (Mm_arr[0] is always 0).
    Q_intp_list : list
        List of Q branch interpolators.
    verbose : bool, optional
        Print verbosity level. Default is 0.
    backend :  ``'gpu-cuda'``, ``'cpu-cuda'``, optional
        Which backend to use; currently only CUDA backends (Nvidia) are supported. ``'cpu-cuda'`` runs the kernel on CPU. Default is ``'gpu-cuda'``.

    Returns
    -------
    init_h : radis.gpu.structs.initData_t
        structue with parameters used for GPU computation that are constant
        during iterations.
    """

    global gpu_mod
    if gpu_mod is not None:
        warn("Only a single GPU context allowed; please call gpu_exit() first.")
        return

    ## First a GPU context is created, then the .ptx file is read
    ## and made available as the GPUModule object gpu_mod
    ## If this fails, None is returned and calculations are
    ## defaulted to CPU emulation

    if backend == "cpu-cuda":
        from radis.gpu.cuda.emulate import CuContext as GPUContext

        ctx = GPUContext.Open(verbose=verbose)
        import radis.gpu.cuda.emulate as backend_module

    else:
        # Try to load GPU
        from radis.gpu.cuda.driver import CuContext as GPUContext

        ctx = GPUContext.Open(verbose=verbose)  # Set verbose to >=2 for comments
        if ctx is None:
            warn(
                NoGPUWarning(
                    "Failed to load CUDA context, this happened either because"
                    + "CUDA is not installed properly, or you have no NVIDIA GPU. "
                    + "Continuing with emulated GPU on CPU..."
                    + "This means *NO* GPU acceleration!"
                )
            )

            # failed to init CUDA context, continue with CPU:
            from radis.gpu.cuda.emulate import CuContext as GPUContext

            ctx = GPUContext.Open(verbose=verbose)
            import radis.gpu.cuda.emulate as backend_module

        else:
            # successfully initialized CUDA context, continue with GPU:
            import radis.gpu.cuda.driver as backend_module

    GPUContext, GPUModule, GPUArray, GPUFFT, GPUTimer = backend_module.getClasses()

    if verbose:
        print("Number of lines loaded: {0}".format(len(v0)))
        print()

    ptx_path = os.path.join(getProjectRoot(), "gpu", "cuda", "build", "kernels.ptx")
    if not os.path.exists(ptx_path):
        raise FileNotFoundError(ptx_path)
    gpu_mod = GPUModule(ctx, ptx_path)  # gpu
    if verbose:
        print("mode:", gpu_mod.getMode())

    ## Next, the GPU is made aware of a number of parameters.
    ## Parameters that don't change during iteration are stored
    ## in init_h. They are copied to the GPU through gpu_mod.setConstant()

    if verbose >= 2:
        print("Copying initialization parameters to device memory...")

    init_h.v_min = vmin
    init_h.dv = dv
    init_h.N_v = Nv
    init_h.N_v_FT = next_fast_len(2 * init_h.N_v)
    init_h.N_x_FT = init_h.N_v_FT // 2 + 1
    init_h.dxG = dxG
    init_h.dxL = dxL
    init_h.N_lines = int(len(v0))
    init_h.N_collision_partners = gamma_arr.shape[0]

    log_c2Mm_arr = np.array(
        [0]
        + [
            0.5 * np.log(8 * k * np.log(2) / (c**2 * Mm * 1e-3 / N_A))
            for Mm in Mm_arr[1:]
        ]
    )
    for i in range(len(log_c2Mm_arr)):
        init_h.log_c2Mm[i] = log_c2Mm_arr[i]

    init_Q(Q_intp_list)
    log_2vMm = np.log(v0) + log_c2Mm_arr.take(iso)

    gpu_mod.setConstant("init_d", init_h)

    init_G_params(log_2vMm.astype(np.float32), verbose)
    init_L_params(na, gamma_arr, verbose)

    if verbose >= 2:
        print("done!")

    ## Next the block- and thread size of the GPU kernels are set.
    ## This determines how the GPU internally divides up the work.

    if verbose >= 2:
        print("Allocating device memory and copying data...")

    NvFT = init_h.N_v_FT
    NxFT = NvFT // 2 + 1
    Ntpb = ctx.getMaxThreadsPerBlock()
    Nli = init_h.N_lines
    threads = (Ntpb, 1, 1)

    gpu_mod.fillLDM.setGrid((Nli // Ntpb + 1, 1, 1), threads)
    gpu_mod.applyLineshapes.setGrid((NxFT // Ntpb + 1, 1, 1), threads)
    gpu_mod.calcTransmittanceNoslit.setGrid((NvFT // Ntpb + 1, 1, 1), threads)
    gpu_mod.applyGaussianSlit.setGrid((NxFT // Ntpb + 1, 1, 1), threads)

    ## Next the variables are initialized on the GPU. Constant variables
    ## that don't change (i.e. pertaining to the database) are immediately
    ## copied to the GPU through GPUArray.fromArray().
    ## Other variables are only allocated. S_klm_d and S_klm_FT_d are
    ## special cases because their shape changes during iteration.
    ## They are not allocated, only given a device pointer by which
    ## they can be referenced later.

    S_klm_d = GPUArray(0, dtype=np.float32, grow_only=True)
    S_klm_FT_d = GPUArray(0, dtype=np.complex64, grow_only=True)

    spectrum_in_d = GPUArray(NxFT, dtype=np.complex64)
    spectrum_out_d = GPUArray(NvFT, dtype=np.float32)

    transmittance_noslit_d = GPUArray(NvFT, dtype=np.float32)
    transmittance_noslit_FT_d = GPUArray(NxFT, dtype=np.complex64)

    transmittance_FT_d = GPUArray(NxFT, dtype=np.complex64)
    transmittance_d = GPUArray(NvFT, dtype=np.float32)

    gpu_mod.fillLDM.setArgs(
        GPUArray.fromArray(iso),
        GPUArray.fromArray(v0),
        GPUArray.fromArray(da),
        GPUArray.fromArray(S0),
        GPUArray.fromArray(El),
        GPUArray.fromArray(gamma_arr),
        GPUArray.fromArray(na),
        S_klm_d,
    )
    gpu_mod.applyLineshapes.setArgs(S_klm_FT_d, spectrum_in_d)
    gpu_mod.calcTransmittanceNoslit.setArgs(spectrum_out_d, transmittance_noslit_d)
    gpu_mod.applyGaussianSlit.setArgs(transmittance_noslit_FT_d, transmittance_FT_d)

    ## FFT's are performed through the GPUFFT object. The required functions are internally
    ## loaded from the cufft library, not through the user kernels (.ptx files).
    ## The FFT's need some memory as "work area". Because the different FFT's can
    ## reuse the work area, we make a GPUArray at this scope that is passed to the
    ## GPUFFT objects. The work area will be scaled according to needs by the GPUFFT objects,
    ## so it can be initialized with a small value.

    workarea_d = GPUArray(0, dtype=np.byte, grow_only=True)
    gpu_mod.fft_fwd = GPUFFT(S_klm_d, S_klm_FT_d, workarea=workarea_d, direction="fwd")
    gpu_mod.fft_rev = GPUFFT(
        spectrum_in_d, spectrum_out_d, workarea=workarea_d, direction="rev"
    )
    gpu_mod.fft_fwd2 = GPUFFT(
        transmittance_noslit_d,
        transmittance_noslit_FT_d,
        workarea=workarea_d,
        direction="fwd",
    )
    gpu_mod.fft_rev2 = GPUFFT(
        transmittance_FT_d, transmittance_d, workarea=workarea_d, direction="rev"
    )

    gpu_mod.timer = GPUTimer()

    if verbose >= 2:
        print("done!")

    return init_h


def gpu_iterate(
    p,
    T,
    mole_fraction,
    verbose=0,
    # for GPU instrument functions (not currently supported):
    l=1.0,
    slit_FWHM=0.0,
):
    """
    Parameters
    ----------
    p : float
        pressure [bar]
    T : float
        temperature [K]
    mole_fraction : float

    Other Parameters
    ----------------
    verbose : int, optional
        The default is 0.


    Returns
    -------
    abscoeff_h : numpy.ndarray[np.float32]
        array with absorbtion coefficients in (cm.-1)
    iter_h : radis.gpu.structs.iterData_t
        structue with parameters used for computation of abscoeff_h.
    times : dict
        dictionary with computation cumulative computation times for
        different stages of the GPU computation. The ``'total'`` key
        gives the total time.
    """

    if gpu_mod is None:
        warn("Must have an open GPU context; please call gpu_init() first.")
        return

    if verbose >= 2:
        print("Copying iteration parameters to device...")

    ## First a number of parameters that change during iteration
    ## are computed and copied to the GPU.

    gpu_mod.timer.reset()

    set_pTQ(p, T, mole_fraction, iter_h, l=l, slit_FWHM=slit_FWHM)
    set_G_params(init_h, iter_h)
    set_L_params(init_h, iter_h)
    gpu_mod.setConstant("iter_d", iter_h)
    gpu_mod.timer.lap("iter_params")

    ## Next the S_klm_d variable is reshaped to the correct shape,
    ## and filled with spectral data.

    if verbose >= 2:
        print("done!")
        print("Filling LDM...")

    S_klm_shape = (init_h.N_v_FT, iter_h.N_G, iter_h.N_L)

    gpu_mod.fillLDM.args[-1].resize(S_klm_shape, init="zeros")
    gpu_mod.fillLDM()
    gpu_mod.timer.lap("fillLDM")

    ## Next the S_klm_FT_d is also reshaped, and the lineshapes are
    ## applied. This consists of an FT of the LDM, a product by the
    ## lineshape FTs & summing all G and L axes, and an inverse FT
    ## on the accumulated spectra.

    if verbose >= 2:
        print("done!")
        print("Applying lineshapes...")

    S_klm_FT_shape = (init_h.N_x_FT, iter_h.N_G, iter_h.N_L)
    gpu_mod.fft_fwd.arr_out.resize(S_klm_FT_shape)
    gpu_mod.fft_fwd()
    gpu_mod.timer.lap("fft_fwd")

    gpu_mod.applyLineshapes()
    gpu_mod.timer.lap("applyLineshapes")

    gpu_mod.fft_rev()
    gpu_mod.timer.lap("fft_rev")

    if verbose >= 2:
        print("Done!")
        print("Calculating transmittance...")

    abscoeff_h = gpu_mod.fft_rev.arr_out.getArray()[: init_h.N_v]

    ## To apply a slit function, first the transmittance is calculated.
    ## Then the convolution is applied by an FT, product with the
    ## instrument function's FT, followed by an inverse FT.

    if verbose >= 2:
        print("Done!")

    ##    ##The code below is to process slits on the GPU, which is currently unsupported.
    ##
    ##        print("Applying slit function...")
    ##
    ##    gpu_mod.calcTransmittanceNoslit()
    ##    gpu_mod.timer.lap("calcTransmittanceNoslit")
    ##
    ##    gpu_mod.fft_fwd2()
    ##    gpu_mod.timer.lap("fft_fwd2")
    ##
    ##    gpu_mod.applyGaussianSlit()
    ##    gpu_mod.timer.lap("applyGaussianSlit")
    ##
    ##    gpu_mod.fft_rev2()
    ##    gpu_mod.timer.lap("fft_rev2")
    ##
    ##    transmittance_h = gpu_mod.fft_rev2.arr_out.getArray()[: init_h.N_v]
    ##
    ##    if verbose >= 2:
    ##        print("done!")

    if verbose == 1:
        print("Finished calculating spectrum!")

    gpu_mod.timer.lap("total")
    times = gpu_mod.timer.getTimes()

    ##    diffs = gpu_mod.timer.getDiffs()
    ##    print(diffs['fillLDM'], diffs['fft_fwd'])

    return abscoeff_h, iter_h, times


def gpu_exit(event=None):
    global gpu_mod
    gpu_mod.context.destroy()
    gpu_mod = None



# -*- coding: utf-8 -*-
"""Common API for RADIS & `Exojax <https://github.com/HajimeKawahara/exojax>`__

-------------------------------------------------------------------------------

"""

from .cdsdapi import cdsd2df
from .hitranapi import hit2df

__all__ = [
    "cdsd2df",
    "hit2df",
]



# -*- coding: utf-8 -*-
"""Created on Wed Oct 18 15:38:14 2017.

@author: erwan

-------------------------------------------------------------------------------
"""

import sys
from time import time


class ProgressBar:
    """Writes completion status and expended time.

    Set it up by creating the object, then calling
    :py:meth:`~radis.misc.progress_bar.ProgressBar.update` and
    :py:meth:`~radis.misc.progress_bar.ProgressBar.done`.

    Parameters
    ----------
    N: int
        (expected) number of iterations
    active: bool
        if ``False``, do not show anything (tip : feed it a ``verbose`` argument)

    Other Parameters
    ----------------
    t0: float
        initializes starting time at ``t0`` (useful for successive loops)

    Example
    -------
    add a progress bar in a loop::

        pb = ProgressBar(N)
        for i, ... in enumerate(...):
            (...)
            pb.update(i)
        pb.done()

    See test in progress_bar.py
    """

    # Todo: One day extend for multiprocss with several progress values?
    # https://stackoverflow.com/questions/7392779/is-it-possible-to-print-a-string-at-a-certain-screen-position-inside-idle

    def __init__(self, N, active=True, t0=None):

        self.t0 = time()
        if t0 is not None:
            self.t0 -= t0
        self.N = N
        self.active = active

    def set_active(self, active=True):
        """Option to activate/deactivate the ProgressBar.

        Used not to make it appear on small processes (based on a
        condition) without changing most of the code
        """
        self.active = active

    def update(self, i, modulo=1, message=""):
        """Update the completion status i/N and time spent.

        Parameters
        ----------
        i: int
            current iteration
        modulo: int
            if higher than ``1``, skip some iterations.
        message: str
            add a custom message. Tip: evaluate your own variables with
            f'{my_variable}' strings

        Example
        -------

        ::


        """
        if not self.active:
            return

        N = self.N
        t0 = self.t0
        if i % modulo == 0:
            if t0 is None:
                msg = "{0:.1f}%\t{1}".format(i / N * 100, message)
            else:
                msg = "({0:.0f}s)\t{1:.1f}%\t{2}".format(
                    time() - t0, i / N * 100, message
                )

            if sys.stdout is not None:
                sys.stdout.write("\r" + msg)
                sys.stdout.flush()

    def done(self):
        """Close the Progress bar."""
        if not self.active:
            return

        self.update(self.N)
        # make new line
        if sys.stdout is not None:
            sys.stdout.write("\n")
            sys.stdout.flush()


# %% Now tested in radis/radis/test/test_misc.py
# def test_progress_bar(*args, **kwargs):
#    ''' Minimal example of a progress bar '''
#
#    from radis.misc.progress_bar import ProgressBar
#    from time import sleep
#    from numpy.random import rand
#
#    print('Testing progress bar')
#
#    a = 0
#    r = list(range(1000))
#    N = len(r)
#    pb = ProgressBar(N)
#    for i in r:
#        pb.update(i,modulo=10)
#        a += i
#        sleep(rand()*3e-3)
#    pb.done()
#
#    return True  # nothing implemented
#
#
# if __name__ == '__main__':
#    test_progress_bar()



import pytest

from radis.api.exomolapi import check_code_level


@pytest.mark.parametrize("bdat_list", [["a1"], ["a0", "a1"], ["a1", "a0"]])
def test_check_bdat_a1(bdat_list):
    bdat = {}
    bdat["code"] = bdat_list
    assert check_code_level(bdat) == "a1"


@pytest.mark.parametrize("bdat_list", [["a0"]])
def test_check_bdat_a0(bdat_list):
    bdat = {}
    bdat["code"] = bdat_list
    assert check_code_level(bdat) == "a0"


@pytest.mark.parametrize("bdat_list", [["a0", "a1", "a2"]])
def test_check_bdat_no_code_level(bdat_list):
    """a2 is not a valid code level"""
    bdat = {}
    bdat["code"] = bdat_list
    assert check_code_level(bdat) == None



# -*- coding: utf-8 -*-
"""
Test Planck functions

-------------------------------------------------------------------------------


"""

import numpy as np
from numpy import pi

from radis.phys.blackbody import planck, planck_wn, sPlanck


def test_exceptions(verbose=True, *args, **kwargs):
    """Here we test that the code actually crashes properly!"""

    import pytest

    # with both wavelength and wavenum
    with pytest.raises(ValueError):  # expected behavior
        sPlanck(
            wavelength_min=300,
            wavelength_max=2000,
            wavenum_min=300,
            wavenum_max=600,
            T=300,
            eps=1,
        )

    # with no temperature
    with pytest.raises(ValueError):  # expected behavior
        sPlanck(wavelength_min=300, wavelength_max=2000, T=None, eps=1)

    # with a wrong epsilon
    with pytest.raises(ValueError):  # expected behavior
        sPlanck(wavelength_min=300, wavelength_max=2000, T=300, eps=10)


def test_planck_nm(verbose=True, plot=True, *args, **kwargs):
    """Test blackbody with Wien's law, Stefan's law and tabulated data
    of maximum

    Reference
    ---------

    Tabulated intensity (75.987...) was checked to match SpectraPlot and
    manually calculated values

    """

    if plot:
        import matplotlib.pyplot as plt

        plt.ion()  # dont get stuck with Matplotlib if executing through pytest

    T = 2200
    eps = 0.36

    s = sPlanck(wavelength_min=300, wavelength_max=8000, T=T, eps=eps, wstep=0.1)

    w_nm = s.get_wavelength(medium="vacuum")
    w_cm = s.get_wavenumber()

    Iunit_per_nm = "W/sr/nm/m2"
    Iunit_per_cm = "W/sr/cm-1/m2"

    if plot:
        s.plot("radiance_noslit", Iunit=Iunit_per_nm)

    I_nm = s.get_radiance_noslit(Iunit=Iunit_per_nm)
    I_cm = s.get_radiance_noslit(Iunit=Iunit_per_cm)

    # Check Wien's law
    w_nm_max = w_nm[I_nm.argmax()]
    assert np.isclose(2898 / T * 1e3, w_nm_max, rtol=1e-4)
    if verbose:
        print("Maximum emission at T={0}K: {1:.2f}nm".format(T, w_nm_max))
        print(".. test_planck_nm.py: Wien's law: OK")

    # Check Stefan's law
    sigma = 5.67e-8  # Stefan-Boltzmann constant   (W/m2/K-4)
    P_theory = eps * sigma * T**4  # Stefan's law (W/m2)
    P_calc = pi * s.get_power("W/m2/sr")  # Lambert's cosine law
    assert np.isclose(P_theory, P_calc, rtol=1e-1)
    if verbose:
        print("Blackbody radiant intensity (Stefan law): {0:.2f} W/m2".format(P_theory))
        print("Blackbody radiant intensity (calculated): {0:.2f} W/m2".format(P_calc))
        print(".. test_planck_nm.py: Stefan's law: OK")

    # Check that max is correct
    # hardcoded for 2200 K, epsilon = 0.36 (~1.3 µm)
    assert I_nm.max() == 75.98736024707178

    # Test planck and planck_wn
    assert np.allclose(planck(w_nm, T, eps, unit=Iunit_per_nm), I_nm)
    assert np.allclose(
        planck_wn(w_cm, T, eps, unit=Iunit_per_cm), I_cm, rtol=1e-2
    )  # higher tolerance because of numerical error
    # during conversion Iunit_per_nm to Iunit_per_cm

    return True


def test_planck_cm(verbose=True, plot=True, *args, **kwargs):
    """Validate Planck calculation with wavenumber

    Notes
    -----

    Earth blackbody radiation [1]_:

    - 294 K
    - eps = 0.65
    - integrated over 2pi steradian


    References
    ----------

    .. [1] Simple Earth Climate Model `NASA Schmidt 2010 <https://www.giss.nasa.gov/research/briefs/schmidt_05/>`__

    """

    if plot:
        import matplotlib.pyplot as plt

        plt.ion()  # dont get stuck with Matplotlib if executing through pytest

    T = 287.2
    eps = 0.78

    s = sPlanck(wavenum_min=10, wavenum_max=3000, T=T, eps=eps, wstep=0.1)

    w_nm = s.get_wavelength()
    w_cm = s.get_wavenumber()

    I_nm = s.get_radiance_noslit(Iunit="mW/sr/m2/nm")
    I_cm = s.get_radiance_noslit(Iunit="mW/sr/m2/cm-1")
    I_cm *= 2 * pi  # mW/m2/cm-1

    # Check Wien's law
    w_nm_max = w_nm[I_nm.argmax()]
    assert np.isclose(2898 / T * 1e3, w_nm_max, rtol=1e-3)
    if verbose:
        print("Maximum emission at T={0}K: {1:.2f}nm".format(T, w_nm_max))
        print(".. test_planck_cm.py: Wien's law: OK")

    # Check Stefan's law
    sigma = 5.67e-8  # Stefan-Boltzmann constant   (W/m2/K-4)
    P_theory = eps * sigma * T**4  # Stefan's law (W/m2)
    P_calc = pi * s.get_power("W/m2/sr")  # Lambert's cosine law
    assert np.isclose(P_theory, P_calc, rtol=1e-3)
    if verbose:
        print(
            "Earth blackbody radiant intensity (Stefan law): {0:.2f} W/m2".format(
                P_theory
            )
        )
        print(
            "Earth blackbody radiant intensity (calculated): {0:.2f} W/m2".format(
                P_calc
            )
        )
        print(".. test_planck_cm.py: Stefan's law: OK")

    if plot:
        plt.figure("test_blackbody.py: test_planck_cm")
        plt.plot(w_cm, I_cm)
        plt.title("Earth blackbody ({0} K, eps={1})".format(T, eps))
        plt.ylabel("Radiance (mW/m2/cm-1)")
        plt.xlabel("Wavenumber (cm-1)")

    return True


def _run_testcases(plot=True, verbose=True, warnings=True, *args, **kwargs):
    """Test procedures

    Parameters
    ----------

    debug: boolean
        swamps the console namespace with local variables. Default ``False``

    """

    # Test all Spectrum methods
    # -------------------------
    test_exceptions()
    test_planck_nm(verbose=verbose, plot=plot, *args, **kwargs)
    test_planck_cm(verbose=verbose, plot=plot, *args, **kwargs)

    return True


if __name__ == "__main__":
    print("test_blackbody:", _run_testcases(verbose=True, plot=True))



# -*- coding: utf-8 -*-
"""
Created on Mon Nov 20 09:59:55 2017

@author: erwan

Examples
--------

Run all tests::

    pytest       (in command line, in project folder)

Run only fast tests (i.e: tests that have a 'fast' label)::

    pytest -m fast

------------------------------------------------------------------------

"""

from os.path import basename

import astropy.units as u
import matplotlib.pyplot as plt
import numpy as np
import pytest

import radis
from radis import config
from radis.lbl.factory import SpectrumFactory
from radis.misc.printer import printm
from radis.test.utils import setup_test_line_databases

fig_prefix = basename(__file__) + ": "

from radis.misc.utils import NotInstalled, not_installed_vaex_args

try:
    import vaex
except ImportError:
    vaex = NotInstalled(*not_installed_vaex_args)

# %% ======================================================================
# TEST
# ---------------------


@pytest.mark.needs_config_file
@pytest.mark.needs_db_CDSD_HITEMP
def test_spec_generation(
    plot=True,
    verbose=2,
    warnings=True,
    update_reference_spectrum=False,
    *args,
    **kwargs
):
    """Test spectrum generation
    Can be used as a base to generate spectra in your codes

    Non-regression test: compare with past version (see conditions below)

    Compare results from a reference case to results calculated on 30/12/2017
    This is not a validation case (30/12/2017 results are not a physically validated
    case), but it makes sure results dont change over time

    Conditions (30/12/2017)::

        Physical Conditions
        ----------------------------------------
           Tgas                 300 K
           Trot                 300 K
           Tvib                 300 K
           pressure             1.01325 bar
           isotope              1,2
           mole_fraction        1
           molecule             CO2
           path_length          1 cm
           wavelength_max       4400.0 nm
           wavelength_min       4150.0 nm
           wavenum_max          2409.6385542168673 cm-1
           wavenum_min          2272.7272727272725 cm-1
        Computation Parameters
        ----------------------------------------
           Tref                 296 K
           broadening_max_width  10 cm-1
           cutoff               1e-25 cm-1/(#.cm-2)
           db_assumed_sorted    True
           db_use_cached        True
           dbformat             cdsd
           dbpath               # USER-DEPENDANT: CDSD-HITEMP
           fillmissinglevelswithzero  False
           levelsfmt            cdsd
           levelspath           # USER-DEPENDANT: CDSD-4000
           medium               vacuum
           parfuncfmt           cdsd
           parfuncpath          # USER-DEPENDANT: CDSD-4000
           rot_distribution     boltzmann
           self_absorption      True
           vib_distribution     boltzmann
           wavenum_max_calc     2414.6385542168673 cm-1
           wavenum_min_calc     2267.7272727272725 cm-1
           waveunit             cm-1
           wstep                0.01 cm-1
        ----------------------------------------

    Notes
    -----

    Performance test. How long it took to calculate this Spectrum?
    Test with cutoff 1e-25, broadening_max_width=10

    - 0.9.15: >>> 33s

    - 0.9.16*: (replaced groupby().apply() with iteration over indexes) >>> 32s
            [but large impact expected on big files]

    - 0.9.16*: (upgraded cache files to h5) >>> 25s

    - 0.9.16*: (also added h5 cache file for levels) >>> 21s

    - 0.9.16*: (with Whiting slit voigt function) >>> 5.8s

    Test with cutoff 1e-27, broadening_max_width=50 :
    ("Spectrum calculated in ... ", including database loading time)

    - 0.9.16*: (same code as last) >>> 12.5s including 7.6s of broadening

    - 0.9.16**: (with pseudo_continuum_threshold=0.01) >>> 7.8s including 2.3s of broadening

    - 0.9.18 (normal code, no pseudo continuum). >>> ?

    - 0.9.21 (normal code) >>> 13.7s, including 8.7s of broadening
             (with pseudo_continuum_threshold=0.01) >>> 4.3s, including 2.6s of broadening

    - 0.9.21*              >>> 14.0s  (added the manual lineshape normalization instead of
                                       Whitings's polynomial)

    - 0.9.22 (normal code) >>> 11.3s   (without energy level lookup, for eq. calculations)
             (with pseudo_continuum_threshold=0.01) >>> 5.9s

    - 0.9.23 (normal code) >>> 7.2s   (added jit in Voigt broadening)
                           >>> 7.1s   (chunksize = None)  (and much faster for more lines)
             (with pseudo_continuum_threshold=0.01) >>> 4.9s

    RADIS:

    - 0.9.19 (normal code) >>> 6.3 s

    - 0.9.20 (normal code) >>> 6.3 s
             (with pseudo_continuum_threshold=0.01) >>> ???
             (with LDM) >>> 2.3 s

    - 0.9.26 (normal code) >>> 7.6 s
             (with pseudo_continuum_threshold=0.01) >>> 2.73s
             (with LDM) >>> 0.25 s

    """

    if plot:  # Make sure matplotlib is interactive so that test are not stuck in pytest
        plt.ion()

    from time import time

    t0 = time()
    if verbose:
        printm(">>> _test_spec_generation")

    # This is how you get a spectrum (see calc.py for front-end functions
    # that do just that)
    sf = SpectrumFactory(
        wavelength_min=4150,
        wavelength_max=4400,
        cutoff=1e-27,
        isotope="1,2",
        truncation=25,
        optimization=None,
        # optimization="min-RMS",
        # pseudo_continuum_threshold=0.01,
        medium="vacuum",
        verbose=verbose,
    )
    sf.warnings["MissingSelfBroadeningWarning"] = "ignore"
    sf.warnings["NegativeEnergiesWarning"] = "ignore"
    sf.load_databank("HITEMP-CO2-DUNHAM")
    s = sf.eq_spectrum(Tgas=300, name="test_spec_generation")
    if verbose:
        printm(
            ">>> _test_spec_generation: Spectrum calculated in {0:.2f}s".format(
                time() - t0
            )
        )

    if plot:
        plt.figure(fig_prefix + "Reference spectrum CDSD-HITEMP (radiance)")
        # Iunit is arbitrary. Use whatever makes sense
        s.plot("radiance_noslit", Iunit="µW/cm2/sr/nm", nfig="same")
    s.rescale_path_length(0.01)

    # Here we get some extra informations:
    if plot:
        sf.plot_broadening(i=0)  # show broadening of one line
        plt.xlim((2267.20, 2268.30))

    # Compare with hardcoded results
    # ... code previously used to export hardcoded results:
    # ... and header contains all input conditions:
    #        np.savetxt('output.txt', np.vstack(s.get('abscoeff', wunit='nm')).T[::10])
    #        print(s)
    # ................
    from radis import get_version
    from radis.test.utils import getTestFile

    wref, Iref = np.loadtxt(getTestFile("CO2abscoeff_300K_4150_4400nm.txt")).T
    match_reference = np.allclose(s.get("abscoeff", wunit="nm")[1][::10], Iref)
    if not match_reference:
        # give some more information before raising error
        printm(
            "Error: {0:.2f}%".format(
                np.mean(abs(s.get("abscoeff", wunit="nm")[1][::10] / Iref - 1)) * 100
            )
        )
        # Store the faulty spectrum
        s.store(
            "test_factory_failed_{0}.spec".format(radis.get_version()),
            if_exists_then="replace",
        )

    # Use "update_reference_spectrum=True" to update reference case :
    if update_reference_spectrum:
        wsave, Isave = s.get("abscoeff", wunit="nm")
        import io
        from contextlib import redirect_stdout

        with io.StringIO() as buf, redirect_stdout(buf):
            print(s)
            s_details = buf.getvalue()
        np.savetxt(
            getTestFile("CO2abscoeff_300K_4150_4400nm.txt"),
            np.vstack((wsave[::10], Isave[::10])).T,
            header="RADIS {0}\n\n{1}".format(
                get_version(add_git_number=False), s_details
            ),
        )

    # Plot comparison
    if plot:
        plt.figure(fig_prefix + "Reference spectrum (abscoeff)")
        # , show_points=True)  # show_points to have an
        s.plot(
            "abscoeff",
            wunit="nm",
            nfig="same",
            lw=3,
            label="RADIS, this version",
        )
        # idea of the resolution
        plt.plot(wref, Iref, "or", ms=3, label="version RADIS 0.9.26 (13/12/20)")
        plt.legend()
        plt.title("All close: {0}".format(match_reference))
        plt.tight_layout()

    # Another example, at higher temperature.
    # Removed because no test is associated with it and it takes time for
    # nothing
    #        s2 = sf.non_eq_spectrum(Tvib=1000, Trot=300)
    #        if plot: s2.plot('abscoeff', wunit='nm')

    if verbose:
        printm(
            "Spectrum calculation (no database loading) took {0:.1f}s\n".format(
                s.conditions["calculation_time"]
            )
        )
        printm("_test_spec_generation finished in {0:.1f}s\n".format(time() - t0))

    assert match_reference


# Test power integral


@pytest.mark.fast
def test_power_integral(verbose=True, warnings=True, *args, **kwargs):
    """Test direct calculation of power integral from Einstein coefficients
    matches integration of broadened spectrum in the optically thin case

    We compare:

    - direct calculation of power integral with equilibrium code
        :meth:`~radis.lbl.SpectrumFactory.optically_thin_power` (T)
    - direct calculation of power integral with non equilibrium code
        :meth:`~radis.lbl.SpectrumFactory.optically_thin_power` (T,T)
    - numerical integration of non equilibrium spectrum under optically thin conditions:
        :meth:`~radis.spectrum.spectrum.Spectrum.get_power`

    Test passes if error < 0.5%

    """

    if verbose:
        printm(">>> _test_power_integral")

    setup_test_line_databases()  # add HITRAN-CO-TEST in ~/radis.json if not there

    sf = SpectrumFactory(
        wavelength_min=4300,
        wavelength_max=4666,
        wstep=0.001,
        cutoff=1e-30,
        path_length=10,
        mole_fraction=400e-6,
        isotope=[1],
        truncation=5,
        verbose=verbose,
    )
    sf.warnings.update(
        {
            "MissingSelfBroadeningWarning": "ignore",
            "OutOfRangeLinesWarning": "ignore",
            "HighTemperatureWarning": "ignore",
        }
    )
    sf.load_databank("HITRAN-CO-TEST", db_use_cached=True)
    unit = "µW/sr/cm2"
    T = 600

    # Calculate:

    # ... direct calculation of power integral with equilibrium code
    Peq = sf.optically_thin_power(Tgas=T, unit=unit)

    # ... direct calculation of power integral with non equilibrium code
    Pneq = sf.optically_thin_power(Tvib=T, Trot=T, unit=unit)

    # ... numerical integration of non equilibrium spectrum under optically thin
    # ... conditions
    sf.input.self_absorption = False
    s = sf.non_eq_spectrum(T, T)

    assert s.conditions["self_absorption"] == False

    # Compare
    err = abs(Peq - s.get_power(unit=unit)) / Peq
    if verbose:
        printm("Emission integral:\t{0:.4g}".format(Peq), unit)
        printm("Emission (noneq code):\t{0:.4g}".format(Pneq), unit)
        printm("Integrated spectrum:\t{0:.4g}".format(s.get_power(unit=unit)), unit)
        printm("Error: {0:.2f}%".format(err * 100))

    assert err < 0.005


@pytest.mark.fast
def test_media_line_shift(plot=False, verbose=True, warnings=True, *args, **kwargs):
    """See wavelength difference in air and vacuum"""

    if plot:  # Make sure matplotlib is interactive so that test are not stuck in pytest
        plt.ion()

    if verbose:
        printm(">>> _test_media_line_shift")

    setup_test_line_databases()  # add HITRAN-CO-TEST in ~/radis.json if not there

    sf = SpectrumFactory(
        wavelength_min=4500,
        wavelength_max=4600,
        wstep=0.001,
        cutoff=1e-30,
        path_length=0.1,
        mole_fraction=400e-6,
        isotope=[1],
        medium="vacuum",
        truncation=5,
        verbose=verbose,
    )
    sf.warnings["MissingSelfBroadeningWarning"] = "ignore"
    sf.warnings["GaussianBroadeningWarning"] = "ignore"
    sf.load_databank("HITRAN-CO-TEST")

    # Calculate a spectrum
    s = sf.eq_spectrum(2000)

    # Compare
    if plot:
        fig = plt.figure(fig_prefix + "Propagating media line shift")
        s.plot("radiance_noslit", wunit="nm_vac", nfig=fig.number, lw=2, label="Vacuum")
        plt.title("CO spectrum (2000 K)")
        s.plot(
            "radiance_noslit",
            wunit="nm",
            nfig=fig.number,
            lw=2,
            color="r",
            label="Air",
        )

    # ... there should be about ~1.25 nm shift at 4.5 µm:
    assert np.isclose(
        s.get("radiance_noslit", wunit="nm_vac")[0][0]
        - s.get("radiance_noslit", wunit="nm")[0][0],
        1.2540436086346745,
    )


@pytest.mark.fast
@pytest.mark.parametrize(
    ("input_wavelengths", "expected_wavelengths_nm"),
    [
        [(4300 * u.nm, 4.5 * u.um), (4300, 4500)],
        [(4500, 5000), (4500, 5000)],
    ],
)
def test_wavelength_units_conversion(
    input_wavelengths, expected_wavelengths_nm, verbose=True, *args, **kwargs
):
    setup_test_line_databases()  # add HITRAN-CO-TEST in ~/radis.json if not there

    wlmin, wlmax = input_wavelengths
    expected_wlmin, expected_wlmax = expected_wavelengths_nm
    sf = SpectrumFactory(
        wavelength_min=wlmin,
        wavelength_max=wlmax,
        wstep=0.01,
        cutoff=1e-30,
        pressure=1,
        path_length=1,
        mole_fraction=1,
        isotope=[1],
        verbose=verbose,
    )
    sf.load_databank("HITRAN-CO-TEST")
    s = sf.eq_spectrum(Tgas=300)
    assert np.isclose(s.get_wavelength().min(), expected_wlmin)
    assert np.isclose(s.get_wavelength().max(), expected_wlmax)


@pytest.mark.fast
@pytest.mark.parametrize(
    ("input_wavenumbers", "expected_wavenumbers_cm1"),
    [
        [(2000 * 1 / u.cm, 230000 * 1 / u.m), (2000, 2300)],
    ],
)
def test_wavenumber_units_conversion(
    input_wavenumbers, expected_wavenumbers_cm1, verbose=True, *args, **kwargs
):
    setup_test_line_databases()  # add HITRAN-CO-TEST in ~/radis.json if not there

    wmin, wmax = input_wavenumbers
    expected_wmin, expected_wmax = expected_wavenumbers_cm1
    sf = SpectrumFactory(
        wavenum_min=wmin,
        wavenum_max=wmax,
        wstep=0.01,
        cutoff=1e-30,
        pressure=1,
        path_length=1,
        mole_fraction=1,
        isotope=[1],
        verbose=verbose,
    )
    sf.load_databank("HITRAN-CO-TEST")
    s = sf.eq_spectrum(Tgas=300)
    assert np.isclose(s.get_wavenumber().min(), expected_wmin)
    assert np.isclose(s.get_wavenumber().max(), expected_wmax)


@pytest.mark.fast
@pytest.mark.parametrize(
    ("input_pressure, expected_pressure_bar"),
    [
        (1, 1),
        (1 * u.mbar, 1e-3),
        (10 * u.bar, 10),
    ],
)
def test_pressure_units_conversion(
    input_pressure, expected_pressure_bar, verbose=True, *args, **kwargs
):
    setup_test_line_databases()  # add HITRAN-CO-TEST in ~/radis.json if not there

    sf = SpectrumFactory(
        wavelength_min=4300,
        wavelength_max=4500,
        wstep=0.01,
        cutoff=1e-30,
        pressure=input_pressure,
        path_length=1,
        mole_fraction=1,
        isotope=[1],
        verbose=verbose,
        warnings={"AccuracyError": "ignore", "AccuracyWarning": "ignore"},
    )
    sf.load_databank("HITRAN-CO-TEST")
    s = sf.eq_spectrum(Tgas=300)
    assert np.isclose(s.conditions["pressure"], expected_pressure_bar)


@pytest.mark.fast
@pytest.mark.parametrize(
    ("input_pathlength, expected_pathlength_cm"),
    [
        (1, 1),
        (1 * u.m, 100),
        (1 * u.cm, 1),
    ],
)
def test_pathlength_units_conversion(
    input_pathlength, expected_pathlength_cm, verbose=True, *args, **kwargs
):
    setup_test_line_databases()  # add HITRAN-CO-TEST in ~/radis.json if not there

    sf = SpectrumFactory(
        wavelength_min=4300,
        wavelength_max=4500,
        wstep=0.01,
        cutoff=1e-30,
        pressure=1,
        path_length=input_pathlength,
        mole_fraction=1,
        isotope=[1],
        verbose=verbose,
    )
    sf.load_databank("HITRAN-CO-TEST")
    s = sf.eq_spectrum(Tgas=300)
    assert np.isclose(s.conditions["path_length"], expected_pathlength_cm)


@pytest.mark.fast
@pytest.mark.parametrize(
    ("input_temperature, expected_temperature_K"),
    [
        (300, 300),
        (300 * u.K, 300),
        (300 * u.deg_C, 573.15),
    ],
)
def test_temperature_units_conversion(
    input_temperature, expected_temperature_K, verbose=True, *args, **kwargs
):
    setup_test_line_databases()  # add HITRAN-CO-TEST in ~/radis.json if not there

    sf = SpectrumFactory(
        wavelength_min=4300,
        wavelength_max=4500,
        wstep=0.001,
        cutoff=1e-30,
        pressure=1,
        mole_fraction=1,
        isotope=[1],
        Tref=300 * u.K,
        verbose=verbose,
    )
    sf.load_databank("HITRAN-CO-TEST")
    s = sf.eq_spectrum(
        Tgas=input_temperature, pressure=20 * u.mbar, path_length=1 * u.mm
    )
    assert np.isclose(s.conditions["Tgas"], expected_temperature_K)
    assert np.isclose(s.conditions["path_length"], 0.1)  # cm
    assert np.isclose(s.conditions["pressure"], 0.02)


@pytest.mark.fast
def test_wstep_auto_method_sf(verbose=True, plot=False, *args, **kwargs):
    """Test to check that on computing several spectrum from the same Spectrum
    Factory object we get the different wstep for each case using auto method"""

    import radis
    from radis.misc.basics import round_off

    setup_test_line_databases()  # add HITRAN-CO-TEST in ~/radis.json if not there

    sf = SpectrumFactory(
        wavelength_min=4400,
        wavelength_max=4800,
        mole_fraction=0.01,
        cutoff=1e-25,
        wstep="auto",
        isotope=[1],
        db_use_cached=True,
        self_absorption=True,
        verbose=verbose,
    )

    sf.warnings["MissingSelfBroadeningWarning"] = "ignore"
    sf.warnings["NegativeEnergiesWarning"] = "ignore"
    sf.warnings["HighTemperatureWarning"] = "ignore"

    sf.load_databank("HITRAN-CO-TEST")

    s1 = sf.eq_spectrum(300, pressure=1)
    assert sf._wstep == "auto"

    wstep_calculated = s1.get_conditions()["wstep"]

    # Checking computed wstep and expected wstep are equal
    assert wstep_calculated == round_off(
        sf.min_width / radis.config["GRIDPOINTS_PER_LINEWIDTH_WARN_THRESHOLD"]
    )

    s2 = sf.eq_spectrum(300, pressure=0.2)
    assert sf._wstep == "auto"

    s3 = sf.eq_spectrum(300, pressure=0.001)
    assert sf._wstep == "auto"

    assert (
        s1.get_conditions()["wstep"]
        != s2.get_conditions()["wstep"]
        != s3.get_conditions()["wstep"]
    )


# @pytest.mark.fast #not fast, Nicolas Minesi 08/04/2024
def test_all_spectrum_using_wstep_auto(verbose=True, plot=False, *args, **kwargs):
    """Checks all methods to calculate Spectrum works with "auto" mode"""
    Tgas = 1000

    sf = SpectrumFactory(
        wavelength_min=4165,
        wavelength_max=4200,
        mole_fraction=1,
        path_length=0.3,
        cutoff=1e-23,
        molecule="CO2",
        isotope=1,
        wstep="auto",
        optimization=None,
        verbose=verbose,
    )
    sf.warnings["MissingSelfBroadeningWarning"] = "ignore"
    sf.warnings["NegativeEnergiesWarning"] = "ignore"
    sf.warnings["HighTemperatureWarning"] = "ignore"
    sf.load_databank("HITRAN-CO2-TEST")

    sf.eq_spectrum(Tgas)
    wstep_1 = sf.params.wstep

    sf.non_eq_spectrum(Tvib=Tgas, Trot=Tgas)
    wstep_2 = sf.params.wstep

    sf.eq_bands(Tgas)
    wstep_3 = sf.params.wstep

    sf.non_eq_bands(Tvib=Tgas, Trot=Tgas)
    wstep_4 = sf.params.wstep

    assert wstep_1 == wstep_2 == wstep_3 == wstep_4
    assert sf._wstep == "auto"


@pytest.mark.skipif(isinstance(vaex, NotInstalled), reason="Vaex not available")
def test_vaex_and_pandas_spectrum():
    """Compares spectrum calculated using vaex and pandas are same.
    CONVOLUTED_QUANTITIES and dataframe df is compared to ensure the spectrum is same.

    Here are things that we ensure are same for both vaex and pandas that
    -Spectra calculated with Vaex & Pandas are the same
    -Spectra calculated with Vaex & Pandas using Database cutoff are the same
    Added in https://github.com/radis/radis/pull/580"""

    from radis import calc_spectrum

    # computing spectrum in vaex dataframe format
    config["DATAFRAME_ENGINE"] = "vaex"
    s_vaex, factory_s_vaex = calc_spectrum(
        1900,
        2300,  # cm-1
        molecule="CO",
        isotope="1,2,3",
        pressure=1.01325,  # bar
        Tgas=700,  # K
        # mole_fraction={'CO': 0.5, 'CO2': 0.3},
        mole_fraction=0.1,
        path_length=1,  # cm
        databank="hitran",  # or 'hitemp', 'geisa', 'exomol'
        # use_cached=False,
        return_factory=True,
    )

    # computing spectrum in pandas dataframe format
    config["DATAFRAME_ENGINE"] = "pandas"
    s_pandas, factory_s_pandas = calc_spectrum(
        1900,
        2300,  # cm-1
        molecule="CO",
        isotope="1,2,3",
        pressure=1.01325,  # bar
        Tgas=700,  # K
        mole_fraction=0.1,
        path_length=1,  # cm
        databank="hitran",  # or 'hitemp', 'geisa', 'exomol'
        # use_cached=False,
        return_factory=True,
    )

    assert np.allclose(
        s_vaex.get("absorbance"), s_pandas.get("absorbance"), equal_nan=True
    )

    for column in factory_s_pandas.df1.columns:
        assert np.all(
            factory_s_vaex.df1[column].to_numpy() == factory_s_pandas.df1[column]
        )

    #%% Additional test with other options (cutoff, ...)
    from radis import SpectrumFactory

    config["DATAFRAME_ENGINE"] = "vaex"
    sf_vaex = SpectrumFactory(
        wavelength_min=4200,
        wavelength_max=4500,
        cutoff=1e-23,
        molecule="CO2",
    )
    sf_vaex.fetch_databank("hitran")
    s_vaex = sf_vaex.eq_spectrum(Tgas=2000)  # failing on the 08/03/2023 - minouHub

    config["DATAFRAME_ENGINE"] = "pandas"
    sf_pd = SpectrumFactory(
        wavelength_min=4200,
        wavelength_max=4500,
        cutoff=1e-23,
        molecule="CO2",
    )
    sf_pd.fetch_databank("hitran")
    s_pd = sf_pd.eq_spectrum(Tgas=2000)
    for column in sf_pd.df1.columns:
        assert np.all(sf_vaex.df1[column].to_numpy() == sf_pd.df1[column])

    assert np.allclose(s_vaex.get("absorbance"), s_pd.get("absorbance"), equal_nan=True)


#%%
@pytest.mark.skipif(isinstance(vaex, NotInstalled), reason="Vaex not available")
def test_vaex_and_pandas_spectrum_noneq():
    """Compares the Spectrum calculated under non-equilibrium conditions .
    Comparison is made between CONVOLUTED_QUANTITIES and dataframe df.

    Here are things that we ensure are same for both vaex and pandas that
    -Spectra calculated with Vaex & Pandas are the same
    -Spectra calculated with Vaex & Pandas using Database cutoff are the same
    Added in https://github.com/radis/radis/pull/580"""

    from radis import calc_spectrum

    conditions = {
        "wmin": 2300,
        "wmax": 2320,
        "molecule": "CO",
        "isotope": 1,
        "pressure": 1.01325,
        "mole_fraction": 0.1,
        "wstep": "auto",
        "path_length": 1,
        "databank": "hitran",
        "verbose": 3,
        "return_factory": True,
    }
    # Calculating spectrum using vae
    config["DATAFRAME_TYPE"] = "vaex"
    s, factory_s = calc_spectrum(
        **conditions,
        Tgas=700,  # K
        Tvib=710,
        Trot=710,
    )

    # Calculating spectrum using pandas
    config["DATAFRAME_ENGINE"] = "pandas"
    s1, factory_s1 = calc_spectrum(
        **conditions,
        Tgas=700,  # K
        Tvib=710,
        Trot=710,
    )

    import numpy as np

    # Comparing different quantities
    assert np.allclose(s.get("absorbance"), s1.get("absorbance"), equal_nan=True)

    # Comparing different columns of dataframe df
    for column in factory_s1.df1.columns:
        assert np.all(factory_s1.df1[column] == factory_s.df1[column].to_numpy())


# --------------------------
if __name__ == "__main__":

    printm("Testing factory:", pytest.main(["test_factory.py", "--pdb"]))
#    printm('Testing factory:', pytest.main(['test_factory.py', '-k', 'test_wavenumber_units_conversion']))



import pytest

from radis.db.classes import is_atom, to_conventional_name


@pytest.mark.fast
def test_is_atom():

    assert is_atom("Fe_I")
    assert is_atom("H_II")
    assert not is_atom("CO2")


@pytest.mark.fast
def test_to_conventional_name():
    assert to_conventional_name("N") == "N_I"
    assert to_conventional_name(801) == "O_II"
    assert to_conventional_name(2700) == "Co_I"
    assert to_conventional_name("H+") == "H_II"
    assert to_conventional_name("Y++++") == "Y_V"
    assert to_conventional_name("26.00") == "Fe_I"
    assert to_conventional_name("06.01") == "C_II"
    assert to_conventional_name("6.01") == "C_II"
    assert to_conventional_name("K_IV") == "K_IV"
    assert to_conventional_name("K IV") == "K_IV"

    # confirm it is friendly to molcules:
    assert to_conventional_name("CO2") == "CO2"
    assert to_conventional_name("OH") == "OH"
    assert to_conventional_name(49) == 49  # ID for COCl2


if __name__ == "__main__":
    test_is_atom()
    test_to_conventional_name()



# -*- coding: utf-8 -*-
"""

Examples
--------

Run all tests::

    pytest      # (in command line, in project folder)

Run only fast tests (i.e: tests that a  'fast' label)::

    pytest -m fast

-------------------------------------------------------------------------------

"""

import numpy as np
import pytest

from radis.los.slabs import MergeSlabs, SerialSlabs


@pytest.mark.fast
def test_merge_slabs(
    verbose=True,
    plot=False,
    close_plots=True,
    warnings=True,
    *args,
    **kwargs,
):
    """Merge 10 slabs with 1/10 of concentration, and compare the results.
    Ensure error is < 0.1%

    Note that we won't have exactly the same results as the broadening is not
    updated in MergeSlabs
    """

    import matplotlib.pyplot as plt

    from radis.test.utils import getTestFile
    from radis.tools.database import load_spec

    if plot:
        plt.ion()  # dont get stuck with Matplotlib if executing through pytest
        if close_plots:
            plt.close("all")

    for optically_thin in [True, False]:

        # Get Some spectra
        s1 = load_spec(getTestFile("CO_Tgas1500K_mole_fraction0.01.spec"), binary=True)
        s2 = load_spec(getTestFile("CO_Tgas1500K_mole_fraction0.5.spec"), binary=True)
        s1.update("all")
        s2.update("all")

        # Merge 50 times s1 in the same slab
        s50 = [s1] * 50
        s1N = MergeSlabs(*s50, optically_thin=optically_thin)

        if plot:
            for k in ["radiance_noslit"]:  # , 'transmittance_noslit']:
                s2.plot(k, lw=3, label="1x[CO=0.5]")
                s1N.plot(k, nfig="same", label="50x[CO=0.01]")
                plt.legend()
                plt.title("Optically thin: {0}".format(optically_thin))
                plt.tight_layout()

        if verbose:
            print("test_merge_slabs")
            print(
                "... Compare 50x[CO=0.01] vs 1x[CON=0.5] (optically thin: {0})".format(
                    optically_thin
                )
            )
            print(
                "... Difference: {0:.2f}%".format(
                    abs(s1N.get_power() / s2.get_power() - 1) * 100
                )
            )
        assert np.isclose(s2.get_power(), s1N.get_power(), 1.5e-2)

    return True


@pytest.mark.fast
def test_equilibrium_condition():
    """See issue #370

    'thermal_equilibrium' is used to recompute some spectral arrays from others,
    it is important to make sure a line-of-sight doesn't assume thermal equilibrium wrongly
    """

    from radis.test.utils import getTestFile
    from radis.tools.database import load_spec

    s1 = load_spec(getTestFile("CO_Tgas1500K_mole_fraction0.01.spec"), binary=True)
    s2 = s1.copy()

    s2.conditions["thermal_equilibrium"] = False
    assert s1.conditions["thermal_equilibrium"] == True
    assert s2.conditions["thermal_equilibrium"] == False

    # Test MergeSlabs
    assert (s1 // s1).is_at_equilibrium() == True
    assert (s2 // s2).is_at_equilibrium() == False
    assert (s1 // s2).is_at_equilibrium() == False
    assert (s2 // s1).is_at_equilibrium() == False

    # Test SerialSlabs
    assert (s1 < s1).is_at_equilibrium() == False
    assert (s2 < s2).is_at_equilibrium() == False
    assert (s1 < s2).is_at_equilibrium() == False
    assert (s2 < s1).is_at_equilibrium() == False


@pytest.mark.fast
def test_serial_slabs_transmittance(
    verbose=True, plot=False, warnings=True, *args, **kwargs
):
    """Add some slabs in series, ensure that overall transmittance decreases
    Also check that some quantities are propagated (like path length?)

    """

    import matplotlib.pyplot as plt

    from radis.test.utils import getTestFile
    from radis.tools.database import load_spec

    if plot:
        plt.ion()  # dont get stuck with Matplotlib if executing through pytest

    # Get Some spectra
    s = load_spec(getTestFile("CO_Tgas1500K_mole_fraction0.5.spec"), binary=True)
    #    s.update('all')
    assert "transmittance_noslit" not in s.get_vars()

    s_los = SerialSlabs(s, s, s)

    w, T = s_los.get("transmittance_noslit")
    # note that it was calculated despite transmittance not in the initial
    # spectrum

    s.update("transmittance_noslit")  # get from abscoeff
    w1, T1 = s.get("transmittance_noslit")

    # Check it looks alright
    assert (T <= T1).all()
    assert (T != T1).any()
    assert (w == w1).all()

    # because they were the same slabs, intensive conditions should have been propagated
    assert s.conditions["Tgas"] == s_los.conditions["Tgas"]
    # but extensive shouldnt:
    assert s.conditions["path_length"] * 3 == s_los.conditions["path_length"]

    if verbose:
        print("Tested Serialslabs transmittance decreases s1>s2: OK")

    return True


@pytest.mark.fast
def test_serial_slabs_radiance(
    verbose=True, plot=False, warnings=True, *args, **kwargs
):
    """Add some slabs in series under optically thin conditions, ensures
    that radiance is about the sum of all.
    """

    import matplotlib.pyplot as plt

    from radis.test.utils import getTestFile
    from radis.tools.database import load_spec

    if plot:
        plt.ion()  # dont get stuck with Matplotlib if executing through pytest

    # Get Some spectra
    s = load_spec(getTestFile("CO_Tgas1500K_mole_fraction0.01.spec"), binary=True)
    s.update("radiance_noslit")

    s_los = SerialSlabs(s, s)

    w, I = s_los.get("radiance_noslit")
    w1, I1 = s.get("radiance_noslit")

    if plot:
        s.plot(nfig="test_serial_slabs_radiance", lw=2)
        plt.title("s>s for almost optically thin conditions")
        s_los.plot(nfig="same")

    # Check it looks alright
    assert np.allclose(I, 2 * I1, rtol=1e-3)

    if verbose:
        print(
            "Tested Serialslabs radiance is added when approximately optically thin: OK"
        )

    # Test operators
    s_serial = s > s
    assert s_serial == s_los

    return True


def _run_testcases(
    verbose=True,
    plot=True,
    close_plots=True,
    warnings=True,
    *args,
    **kwargs,
):

    test_merge_slabs(
        verbose=verbose,
        plot=plot,
        close_plots=close_plots,
        warnings=warnings,
        *args,
        **kwargs,
    )

    test_serial_slabs_transmittance(
        verbose=verbose, plot=plot, warnings=warnings, *args, **kwargs
    )

    test_serial_slabs_radiance(
        verbose=verbose, plot=plot, warnings=warnings, *args, **kwargs
    )

    test_equilibrium_condition()

    return True


if __name__ == "__main__":
    print("test_slabs: ", _run_testcases(verbose=True))



# -*- coding: utf-8 -*-
"""

-------------------------------------------------------------------------------


"""

import warnings


def Unit(st, *args, **kwargs):
    """Radis evaluation of an unit, using :py:class:`~astropy.units.Unit`

    Changes compare to Astropy standards:

    - "µm" is accepted and converted to "um"
    - "/molecule" or "/molec" are removed, but do not raise an error
    - we do not raise a warning if multiple slashes, e.g. "mW/cm2/sr/nm"

    Examples
    --------

    ::

        from radis.phys.units import Unit as u
        a = 200 * u("mW/cm2/sr/nm")
        a += 0.1 * u("W/cm2/sr/nm")
    """

    import astropy.units as u

    try:
        st = st.replace("µ", "u")
        st = st.replace("/molecule", "")
        st = st.replace("/molec", "")
    except AttributeError:
        pass

    with warnings.catch_warnings():
        # Ignore some warnings : Cause all warnings to always be triggered.
        warnings.filterwarnings(
            "ignore", ".*multiple slashes.*", category=u.UnitsWarning
        )
        unit = u.Unit(st, *args, **kwargs)
    return unit


def conv2(quantity, fromunit, tounit):
    """Converts `quantity` from unit `fromunit` to unit `tounit`

    Parameters
    ----------
    quantity: array
        quantity to convert
    fromunit: str
        input unit
    tounit: str
        output unit

    Note
    ----

    1.

    The output is still non dimensional. We don't transform `quantity`
    into a dimensioned array because this may create a performance
    drop in computationally-expensive task. Instead, we assume we know for
    sure the units in which some of our quantities will be created, and just
    want to let the users choose another output unit
    """

    if fromunit == tounit:
        return quantity
    else:

        import astropy.units as u

        try:
            a = quantity * Unit(fromunit)
            a = a.to(Unit(tounit))

        except u.UnitConversionError as err:
            suggestion = ""
            if fromunit == "" or tounit == "":
                suggestion += " Is one of the two arrays normalized ?"
            raise TypeError(
                f"Cannot convert quantity `{fromunit}` to the specified unit `{tounit}`. Please check the dimensions."
                + suggestion
            ) from err

        return a.value


def is_homogeneous(unit1, unit2):
    """Tells if unit1 and unit2 are homogeneous, using the Astropy library.

    Parameters
    ----------
    unit1, unit2: str
        units
    """

    if unit1 == unit2:
        return True
    else:

        import astropy.units as u

        try:
            1 * Unit(unit1) + 1 * Unit(unit2)
            return True
        except u.UnitConversionError:
            return False


# %% Emission density conversion functions (with changes from ~1/nm in ~1/cm-1)


def convert_emi2cm(j_nm, wavenum, Iunit0, Iunit):
    """Convert spectral emission density in wavelength base (typically
    ~mW/cm3/sr/nm) to spectral emission density in wavenumber base
    (~mW/cm3/sr/cm-1)

    Parameters
    ----------
    j_nm: array
        spectral emission density in Iunit0 unit (~ per wavelength) )
    wavenum: array (cm-1)
        wavenumber
    Iunit0: str
        unit (~ per wavelength) to convert from
    Iunit: str
        unit (~ per wavenumber) to convert to


    Notes
    -----
    Implementation:

    We use the variable substitution::

            dλ/dν = - 1/ν**2        [in SI]
            dλ/dν = - 10^7 / ν**2   [nm -> cm-1]

    We want   Lν.dν = Lλ.dλ
    so        Lν = - Lλ * 1e7 / ν^2


    Example
    -------
    Validation::

        w_nm, j_nm = s.get('emisscoeff', 'nm', 'mW/cm2/sr/nm')
        w_cm, j_cm = s.get('emisscoeff', 'cm', 'mW/cm2/sr/cm-1')
        print(trapz(y_nm, x_nm))
        print(trapz(y_cm, x_cm))

    Both integrals are to be the same
    """

    j_nm = conv2(j_nm, Iunit0, "mW/cm**3/sr/nm")

    # Convert ''mW/cm3/sr/nm' to 'mW/cm3/sr/cm-1'
    j_cm = j_nm * 1e7 / wavenum**2
    # note that we discard the - sign.
    # This means that one of the integrals `trapz(j_nm, wavelen)` or
    # `trapz(j_cm, wavenum)` will be < 0 as they are sorted in opposite order.
    # But both plots look alright

    # Convert to whatever was wanted
    j_cm = conv2(j_cm, "mW/cm3/sr/cm-1", Iunit)

    return j_cm


def convert_emi2nm(j_cm, wavenum, Iunit0, Iunit):
    """Convert spectral emission density in wavenumber base (typically
    ~mW/cm3/sr/cm-1) to spectral radiance in wavelength base (~mW/cm3/sr/nm)

    Parameters
    ----------
    j_cm: array
        spectral emission density in Iunit0 unit (~ per wavenumber)
    wavenum: array (cm-1)
        wavenumber
    Iunit0: str
        unit (~ per wavenumber) to convert from
    Iunit: str
        unit (~ per wavelength) to convert to


    Notes
    -----
    Implementation:

    We use the variable substitution::

            dλ/dν = - 1/ν**2        [in SI]
            dλ/dν = - 10^7 / ν**2   [nm -> cm-1]

    We want   Lν.dν = Lλ.dλ
    so        Lλ  = - Lν * 1e-7 * ν^2
    """

    j_cm = conv2(j_cm, Iunit0, "mW/cm3/sr/cm-1")

    # Convert 'mW/cm3/sr/cm-1' to 'mW/cm3/sr/nm'
    j_nm = j_cm * 1e-7 * wavenum**2
    # note that we discard the - sign.
    # This means that one of the integrals `trapz(L_nm, wavelen)` or
    # `trapz(L_cm, wavenum)` will be < 0 as they are sorted in opposite order.
    # But both plots look alright

    # Convert to whatever was wanted
    j_nm = conv2(j_nm, "mW/cm3/sr/nm", Iunit)

    return j_nm


# %% Radiance conversion functions (with changes from ~1/nm in ~1/cm-1)


def convert_rad2cm(
    l_nm,
    wavenum,
    Iunit0,
    Iunit,
    per_nm_is_like="mW/sr/cm2/nm",
    per_cm_is_like="mW/sr/cm2/cm-1",
):
    """Convert spectral radiance in wavelength base (~1/nm) to spectral
    radiance in wavenumber base (~1/cm-1)

    Parameters
    ----------
    l_nm: array
        spectral radiance in yunit0 unit (~ per wavelength)
    wavenum: array (cm-1)
        wavenumber
    Iunit0: str
        unit (~ per wavelength) to convert from
    Iunit: str
        unit (~ per wavenumber) to convert to


    Notes
    -----
    Implementation:

    We use the variable substitution:
            dλ/dν = - 1/ν**2        [in SI]
            dλ/dν = - 10^7 / ν**2   [nm -> cm-1]

    We want   Lν.dν = Lλ.dλ
    so        Lν = - Lλ * 1e7 / ν^2


    Example
    -------
    Validation::

        x_nm, y_nm = s.get('radiance_noslit', 'nm', 'mW/cm2/sr/nm')
        x_cm, y_cm = s.get('radiance_noslit', 'cm', 'mW/cm2/sr/cm-1')
        print(trapz(y_nm, x_nm))
        print(trapz(y_cm, x_cm))

    Both integrals are to be the same
    """

    l_nm = conv2(l_nm, Iunit0, per_nm_is_like)

    # Convert ''mW/cm2/sr/nm' to 'mW/cm2/sr/cm-1'
    l_cm = l_nm * 1e7 / wavenum**2
    # note that we discard the - sign.
    # This means that one of the integrals `trapz(L_nm, wavelen)` or
    # `trapz(L_cm, wavenum)` will be < 0 as they are sorted in opposite order.
    # But both plots look alright

    # Convert to whatever was wanted
    l_cm = conv2(l_cm, per_cm_is_like, Iunit)

    return l_cm


def convert_rad2nm(
    l_cm,
    wavenum,
    Iunit0,
    Iunit,
    per_nm_is_like="mW/sr/cm2/nm",
    per_cm_is_like="mW/sr/cm2/cm-1",
):
    """Convert spectral radiance in wavenumber base (~1/cm-1) to spectral
    radiance in wavelength base (~1/nm)

    Parameters
    ----------
    l_cm: array
        spectral radiance in yunit0 unit (~ per wavenumber)
    wavenum: array (cm-1)
        wavenumber
    Iunit0: str
        unit (~ per wavenumber) to convert from
    Iunit: str
        unit (~ per wavelength) to convert to

    Notes
    -----
    Implementation:

    We use a variable substitution::

            dλ/dν = - 1/ν**2        [in SI]
            dλ/dν = - 10^7 / ν**2   [nm -> cm-1]

    We want   Lν.dν = Lλ.dλ
    so        Lλ  = - Lν * 1e-7 * ν^2
    """

    l_cm = conv2(l_cm, Iunit0, per_cm_is_like)

    # Convert 'mW/cm2/sr/cm-1' to 'mW/cm2/sr/nm'
    l_nm = l_cm * 1e-7 * wavenum**2
    # note that we discard the - sign.
    # This means that one of the integrals `trapz(L_nm, wavelen)` or
    # `trapz(L_cm, wavenum)` will be < 0 as they are sorted in opposite order.
    # But both plots look alright

    # Convert to whatever was wanted
    l_nm = conv2(l_nm, per_nm_is_like, Iunit)

    return l_nm


def convert_universal(
    I,
    from_unit,
    to_unit,
    wavenum,
    per_nm_is_like="mW/cm2/sr/nm",
    per_cm_is_like="mW/cm2/sr/cm-1",
):
    """Return variable var in whatever unit, and converts to to_unit Also deal
    with cases where var is in ~1/nm (per_nm_is_like) or ~1/cm-1
    (per_cm_is_like)

    Parameters
    ----------
    var: str
        variable to get. Usually 'radiance' or 'radiance_noslit'
    to_unit: str
        unit to convert variable to
    wavenum: array (cm-1)
        wavenumber is needed in case we convert from ~1/nm to ~1/cm-1 (requires
        a change of variable in the integral)
    """

    if from_unit == to_unit:
        return I
    else:

        import astropy.units as u

        Iunit0 = from_unit
        Iunit = to_unit
        try:
            if is_homogeneous(Iunit0, per_nm_is_like) and is_homogeneous(
                Iunit, per_cm_is_like
            ):
                I = convert_rad2cm(
                    I,
                    wavenum,
                    Iunit0,
                    Iunit,
                    per_nm_is_like=per_nm_is_like,
                    per_cm_is_like=per_cm_is_like,
                )
                # note that there may still be a new DimensionalityError
                # raise here if the input was non-sense.
            elif is_homogeneous(Iunit0, per_cm_is_like) and is_homogeneous(
                Iunit, per_nm_is_like
            ):
                I = convert_rad2nm(
                    I,
                    wavenum,
                    Iunit0,
                    Iunit,
                    per_nm_is_like=per_nm_is_like,
                    per_cm_is_like=per_cm_is_like,
                )
                # note that there may still be a new DimensionalityError
                # raise here if the input was non-sense.
            else:  # general case: just convert
                I = conv2(I, Iunit0, Iunit)
        except u.UnitConversionError:
            raise TypeError(Iunit0, Iunit)

        return I


# %% Test
if __name__ == "__main__":
    from radis.test.test_phys import test_units

    print(("Test :", test_units()))
